"""A Fortran 90 namelist parser and generator.

:copyright: Copyright 2014 Marshall Ward, see AUTHORS for details.
:license: Apache License, Version 2.0, see LICENSE for details.
"""
from f90nml.namelist import Namelist
from f90nml.parser import Parser

__version__ = '1.3.1'


def read(nml_path):
    """Parse a Fortran namelist file and return its contents.

    File object usage:

    >>> with open(nml_path) as nml_file:
    >>>     nml = f90nml.read(nml_file)

    File path usage:

    >>> nml = f90nml.read(nml_path)

    This function is equivalent to the ``read`` function of the ``Parser``
    object.

    >>> parser = f90nml.Parser()
    >>> nml = parser.read(nml_file)
    """
    parser = Parser()
    return parser.read(nml_path)


def reads(nml_string):
    """Parse a Fortran namelist string and return its contents.

    >>> nml_str = '&data_nml x=1 y=2 /'
    >>> nml = f90nml.reads(nml_str)

    This function is equivalent to the ``reads`` function of the ``Parser``
    object.

    >>> parser = f90nml.Parser()
    >>> nml = parser.reads(nml_str)
    """
    parser = Parser()
    return parser.reads(nml_string)


def write(nml, nml_path, force=False, sort=False):
    """Save a namelist to disk using either a file object or its file path.

    File object usage:

    >>> with open(nml_path, 'w') as nml_file:
    >>>     f90nml.write(nml, nml_file)

    File path usage:

    >>> f90nml.write(nml, 'data.nml')

    This function is equivalent to the ``write`` function of the ``Namelist``
    object ``nml``.

    >>> nml.write('data.nml')

    By default, ``write`` will not overwrite an existing file.  To override
    this, use the ``force`` flag.

    >>> nml.write('data.nml', force=True)

    To alphabetically sort the ``Namelist`` keys, use the ``sort`` flag.

    >>> nml.write('data.nml', sort=True)
    """
    # Promote dicts to Namelists
    if not isinstance(nml, Namelist) and isinstance(nml, dict):
        nml_in = Namelist(nml)
    else:
        nml_in = nml

    nml_in.write(nml_path, force=force, sort=sort)


def patch(nml_path, nml_patch, out_path=None):
    """Create a new namelist based on an input namelist and reference dict.

    >>> f90nml.patch('data.nml', nml_patch, 'patched_data.nml')

    This function is equivalent to the ``read`` function of the ``Parser``
    object with the patch output arguments.

    >>> parser = f90nml.Parser()
    >>> nml = parser.read('data.nml', nml_patch, 'patched_data.nml')

    A patched namelist file will retain any formatting or comments from the
    original namelist file.  Any modified values will be formatted based on the
    settings of the ``Namelist`` object.
    """
    parser = Parser()

    return parser.read(nml_path, nml_patch, out_path)



"""The f90nml namelist parser.

The ``Parser`` object converts the contents of a Fortran namelist into a
hierarchy of Python dicts containing equivalent intrinsic Python data types.

:copyright: Copyright 2014 Marshall Ward, see AUTHORS for details.
:license: Apache License, Version 2.0, see LICENSE for details.
"""
from __future__ import print_function

import warnings
import copy
from string import whitespace
import itertools

from f90nml.findex import FIndex
from f90nml.fpy import pyfloat, pycomplex, pybool, pystr
from f90nml.namelist import Namelist
from f90nml.tokenizer import Tokenizer


class Parser(object):
    """Fortran namelist parser."""

    def __init__(self):
        """Create the parser object."""
        # Token management
        self.tokens = None
        self.token = None
        self.prior_token = None

        # Patching
        self.pfile = None

        # Configuration
        self._default_start_index = 1
        self._global_start_index = None
        self._comment_tokens = '!'
        self._sparse_arrays = False
        self._row_major = False
        self._strict_logical = True

    @property
    def comment_tokens(self):
        """Return a string of single-character comment tokens in the namelist.

        :type: ``str``
        :default: ``'!'``

        Some Fortran programs will introduce alternative comment tokens (e.g.
        ``#``) for internal preprocessing.

        If you need to support these tokens, create a ``Parser`` object and set
        the comment token as follows:

        >>> parser = f90nml.Parser()
        >>> parser.comment_tokens += '#'
        >>> nml = parser.read('sample.nml')

        Be aware that this is non-standard Fortran and could mangle any strings
        using the ``#`` characters.  Characters inside string delimiters should
        be protected, however.
        """
        return self._comment_tokens

    @comment_tokens.setter
    def comment_tokens(self, value):
        """Validate and set the comment token string."""
        if not isinstance(value, str):
            raise TypeError('comment_tokens attribute must be a string.')
        self._comment_tokens = value

    @property
    def default_start_index(self):
        """Assumed starting index for a vector.

        :type: ``int``
        :default: 1

        Since Fortran allows users to set an arbitrary start index, it is not
        always possible to assign an index to values when no index range has
        been provided.

        For example, in the namelist ``idx.nml`` shown below, the index of the
        values in the second assignment are ambiguous and depend on the
        implicit starting index.

        .. code-block:: fortran

           &idx_nml
               v(3:5) = 3, 4, 5
               v = 1, 2
           /

        The indices of the second entry in ``v`` are ambiguous.  The result for
        different values of ``default_start_index`` are shown below.

        >>> parser = f90nml.Parser()
        >>> parser.default_start_index = 1
        >>> nml = parser.read('idx.nml')
        >>> nml['idx_nml']['v']
        [1, 2, 3, 4, 5]

        >>> parser.default_start_index = 0
        >>> nml = parser.read('idx.nml')
        >>> nml['idx_nml']['v']
        [1, 2, None, 3, 4, 5]
        """
        return self._default_start_index

    @default_start_index.setter
    def default_start_index(self, value):
        """Validate and set the default start index."""
        if not isinstance(value, int):
            raise TypeError('default_start_index attribute must be of int '
                            'type.')
        self._default_start_index = value

    @property
    def global_start_index(self):
        """Define an explicit start index for all vectors.

        :type: ``int``, ``None``
        :default: ``None``

        When set to ``None``, vectors are assumed to start at the lowest
        specified index.  If no index appears in the namelist, then
        ``default_start_index`` is used.

        When ``global_start_index`` is set, then all vectors will be created
        using this starting index.

        For the namelist file ``idx.nml`` shown below,

        .. code-block:: fortran

           &idx_nml
              v(3:5) = 3, 4, 5
           /

        the following Python code behaves as shown below.

        >>> parser = f90nml.Parser()
        >>> nml = parser.read('idx.nml')
        >>> nml['idx_nml']['v']
        [3, 4, 5]

        >>> parser.global_start_index = 1
        >>> nml = parser.read('idx.nml')
        >>> nml['idx_nml']['v']
        [None, None, 3, 4, 5]

        Currently, this property expects a scalar, and applies this value to
        all dimensions.
        """
        return self._global_start_index

    @global_start_index.setter
    def global_start_index(self, value):
        """Set the global start index."""
        if not isinstance(value, int) and value is not None:
            raise TypeError('global_start_index attribute must be of int '
                            'type.')
        self._global_start_index = value

    @property
    def row_major(self):
        """Read multidimensional arrays in row-major format.

        :type: ``bool``
        :default: ``False``

        Multidimensional array data contiguity is preserved by default, so that
        column-major Fortran data is represented as row-major Python list of
        lists.

        The ``row_major`` flag will reorder the data to preserve the index
        rules between Fortran to Python, but the data will be converted to
        row-major form (with respect to Fortran).
        """
        return self._row_major

    @row_major.setter
    def row_major(self, value):
        """Validate and set row-major format for multidimensional arrays."""
        if value is not None:
            if not isinstance(value, bool):
                raise TypeError(
                    'f90nml: error: row_major must be a logical value.')
            else:
                self._row_major = value

    @property
    def sparse_arrays(self):
        """Store unset rows of multidimensional arrays as empty lists.

        :type: ``bool``
        :default: ``False``

        Enabling this flag will replace rows of unset values with empty lists,
        and will also not pad any existing rows when other rows are expanded.

        This is not a true sparse representation, but rather is slightly more
        sparse than the default dense array representation.
        """
        return self._sparse_arrays

    @sparse_arrays.setter
    def sparse_arrays(self, value):
        """Validate and enable spare arrays."""
        if not isinstance(value, bool):
            raise TypeError('sparse_arrays attribute must be a logical type.')
        self._sparse_arrays = value

    @property
    def strict_logical(self):
        """Use strict rules for parsing logical data value parsing.

        :type: ``bool``
        :default: ``True``

        The ``strict_logical`` flag will limit the parsing of non-delimited
        logical strings as logical values.  The default value is ``True``.

        When ``strict_logical`` is enabled, only ``.true.``, ``.t.``, ``true``,
        and ``t`` are interpreted as ``True``, and only ``.false.``, ``.f.``,
        ``false``, and ``f`` are interpreted as false.

        When ``strict_logical`` is disabled, any value starting with ``.t`` or
        ``t`` is interpreted as ``True``, while any string starting with ``.f``
        or ``f`` is interpreted as ``False``, as described in the language
        standard.  However, it can interfere with namelists which contain
        non-delimited strings.
        """
        return self._strict_logical

    @strict_logical.setter
    def strict_logical(self, value):
        """Validate and set the strict logical flag."""
        if value is not None:
            if not isinstance(value, bool):
                raise TypeError(
                    'f90nml: error: strict_logical must be a logical value.')
            else:
                self._strict_logical = value

    def read(self, nml_fname, nml_patch_in=None, patch_fname=None):
        """Parse a Fortran namelist file and store the contents.

        >>> parser = f90nml.Parser()
        >>> data_nml = parser.read('data.nml')
        """
        # For switching based on files versus paths
        nml_is_path = not hasattr(nml_fname, 'read')
        patch_is_path = not hasattr(patch_fname, 'read')

        # Convert patch data to a Namelist object
        if nml_patch_in is not None:
            if not isinstance(nml_patch_in, dict):
                raise TypeError('Input patch must be a dict or a Namelist.')

            nml_patch = copy.deepcopy(Namelist(nml_patch_in))

            if not patch_fname and nml_is_path:
                patch_fname = nml_fname + '~'
            elif not patch_fname:
                raise ValueError('f90nml: error: No output file for patch.')
            elif nml_fname == patch_fname:
                raise ValueError('f90nml: error: Patch filepath cannot be the '
                                 'same as the original filepath.')
            if patch_is_path:
                self.pfile = open(patch_fname, 'w')
            else:
                self.pfile = patch_fname
        else:
            nml_patch = Namelist()

        try:
            nml_file = open(nml_fname, 'r') if nml_is_path else nml_fname
            try:
                return self._readstream(nml_file, nml_patch)
            except StopIteration:
                raise ValueError('End-of-file reached before end of namelist.')

            # Close the files we opened on any exceptions within readstream
            finally:
                if nml_is_path:
                    nml_file.close()
        finally:
            if self.pfile and patch_is_path:
                self.pfile.close()

    def reads(self, nml_string):
        """Parse a namelist string and return an equivalent Namelist object.

        >>> parser = f90nml.Parser()
        >>> data_nml = parser.reads('&data_nml x=1 y=2 /')
        """
        try:
            return self._readstream(iter(nml_string.splitlines(True)))
        except StopIteration:
            raise ValueError('End-of-file reached before end of namelist.')

    def _readstream(self, nml_file, nml_patch_in=None):
        """Parse an input stream containing a Fortran namelist."""
        nml_patch = nml_patch_in if nml_patch_in is not None else Namelist()

        tokenizer = Tokenizer()
        tokenizer.comment_tokens = self.comment_tokens
        f90lex = []
        for line in nml_file:
            toks = tokenizer.parse(line)
            while tokenizer.prior_delim:
                new_toks = tokenizer.parse(next(nml_file))

                # Skip empty lines
                if not new_toks:
                    continue

                # The tokenizer always pre-tokenizes the whitespace (leftover
                # behaviour from Fortran source parsing) so this must be added
                # manually.
                if new_toks[0].isspace():
                    toks[-1] += new_toks.pop(0)

                # Append the rest of the string (if present)
                if new_toks:
                    toks[-1] += new_toks[0]

                    # Attach the rest of the tokens
                    toks.extend(new_toks[1:])

            toks.append('\n')
            f90lex.extend(toks)

        self.tokens = iter(f90lex)

        nmls = Namelist()

        # Attempt to get first token; abort on empty file
        try:
            self._update_tokens(write_token=False)
        except StopIteration:
            return nmls

        # TODO: Replace "while True" with an update_token() iterator
        while True:
            try:
                # Check for classic group terminator
                if self.token == 'end':
                    self._update_tokens()

                # Ignore tokens outside of namelist groups
                while self.token not in ('&', '$'):
                    self._update_tokens()

            except StopIteration:
                break

            # Create the next namelist
            try:
                self._update_tokens()
            except StopIteration:
                raise ValueError('End-of-file after namelist group token `&`.')
            g_name = self.token

            g_vars = Namelist()
            v_name = None

            # TODO: Edit `Namelist` to support case-insensitive `get` calls
            grp_patch = nml_patch.pop(g_name.lower(), Namelist())

            # Populate the namelist group
            while g_name:

                if self.token not in ('=', '%', '('):
                    try:
                        self._update_tokens()
                    except StopIteration:
                        raise ValueError(
                            'End-of-file before end of namelist group: \'&{}\''
                            ''.format(g_name)
                        )

                # Set the next active variable
                if self.token in ('=', '(', '%'):

                    v_name, v_values = self._parse_variable(
                        g_vars,
                        patch_nml=grp_patch
                    )

                    if v_name in g_vars:
                        v_prior_values = g_vars[v_name]
                        v_values = merge_values(v_prior_values, v_values)

                    g_vars[v_name] = v_values

                    # Deselect variable
                    v_name = None
                    v_values = []

                # Finalise namelist group
                if self.token in ('/', '&', '$'):

                    # Append any remaining patched variables
                    for v_name, v_val in grp_patch.items():
                        g_vars[v_name] = v_val
                        v_strs = nmls._var_strings(v_name, v_val)
                        for v_str in v_strs:
                            self.pfile.write(v_str + '\n')

                    # Append the grouplist to the namelist
                    if g_name in nmls:
                        g_update = nmls[g_name]

                        # Update to list of groups
                        if not isinstance(g_update, list):
                            g_update = [g_update]

                        g_update.append(g_vars)

                    else:
                        g_update = g_vars

                    nmls[g_name] = g_update

                    # Reset state
                    g_name, g_vars = None, None

            try:
                self._update_tokens()
            except StopIteration:
                break

        if nml_patch:
            # Append the contents to the namelist patch
            print(file=self.pfile)
            print(nml_patch, file=self.pfile)

            # Now append the values to the output namelist
            for grp in nml_patch:
                nmls[grp] = nml_patch[grp]

        return nmls

    def _parse_variable(self, parent, patch_nml=None):
        """Parse a variable and return its name and values."""
        if not patch_nml:
            patch_nml = Namelist()

        v_name = self.prior_token
        v_values = []

        # Patch state
        patch_values = None

        # Derived type parent index (see notes below)
        dt_idx = None

        if self.token == '(':

            v_idx_bounds = self._parse_indices()
            v_idx = FIndex(v_idx_bounds, self.global_start_index)

            # Update starting index against namelist record
            if v_name.lower() in parent.start_index:
                p_idx = parent.start_index[v_name.lower()]

                for idx, pv in enumerate(zip(p_idx, v_idx.first)):
                    if all(i is None for i in pv):
                        i_first = None
                    else:
                        i_first = min(i for i in pv if i is not None)

                    v_idx.first[idx] = i_first

                # Resize vector based on starting index
                parent[v_name] = prepad_array(parent[v_name], p_idx,
                                              v_idx.first)
            else:
                # If variable already existed without an index, then assume a
                #   1-based index
                # FIXME: Need to respect undefined `None` starting indexes?
                if v_name in parent:
                    v_idx.first = [self.default_start_index
                                   for _ in v_idx.first]

            parent.start_index[v_name.lower()] = v_idx.first

            self._update_tokens()

            # Derived type parent check
            # NOTE: This assumes single-dimension derived type vectors
            #       (which I think is the only case supported in Fortran)
            if self.token == '%':
                assert v_idx_bounds[0][1] - v_idx_bounds[0][0] == 1
                dt_idx = v_idx_bounds[0][0] - v_idx.first[0]

                # NOTE: This is the sensible play to call `parse_variable`
                # but not yet sure how to implement it, so we currently pass
                # along `dt_idx` to the `%` handler.

        else:
            v_idx = None

            # If indexed variable already exists, then re-index this new
            #   non-indexed variable using the global start index

            if v_name in parent.start_index:
                p_start = parent.start_index[v_name.lower()]
                v_start = [self.default_start_index for _ in p_start]

                # Resize vector based on new starting index
                for i_p, i_v in zip(p_start, v_start):
                    if i_v < i_p:
                        pad = [None for _ in range(i_p - i_v)]
                        parent[v_name] = pad + parent[v_name]

                parent.start_index[v_name.lower()] = v_start

        if self.token == '%':

            # Resolve the derived type

            # Check for value in patch
            v_patch_nml = None
            if v_name in patch_nml:
                v_patch_nml = patch_nml.pop(v_name.lower())

            if parent:
                vpar = parent.get(v_name.lower())
                if vpar and isinstance(vpar, list):
                    # If new element is not a list, then assume it's the first
                    # element of the list.
                    if dt_idx is None:
                        dt_idx = self.default_start_index

                    try:
                        v_parent = vpar[dt_idx]
                    except IndexError:
                        v_parent = Namelist()
                elif vpar:
                    v_parent = vpar
                else:
                    v_parent = Namelist()
            else:
                v_parent = Namelist()
                parent[v_name] = v_parent

            self._update_tokens()
            self._update_tokens()

            v_att, v_att_vals = self._parse_variable(
                v_parent,
                patch_nml=v_patch_nml
            )

            next_value = Namelist()
            next_value[v_att] = v_att_vals
            self._append_value(v_values, next_value, v_idx)

        else:
            # Construct the variable array

            assert self.token == '='
            n_vals = None

            self._update_tokens()

            # Check if value is in the namelist patch
            # TODO: Edit `Namelist` to support case-insensitive `pop` calls
            #       (Currently only a problem in PyPy2)
            if v_name in patch_nml:
                patch_values = patch_nml.pop(v_name.lower())

                if not isinstance(patch_values, list):
                    patch_values = [patch_values]

                p_idx = 0

            # Add variables until next variable trigger
            while (self.token not in ('=', '(', '%') or
                   (self.prior_token, self.token) in (('=', '('), (',', '('))):

                # Check for repeated values
                if self.token == '*':
                    n_vals = self._parse_value()
                    assert isinstance(n_vals, int)
                    self._update_tokens()
                elif not n_vals:
                    n_vals = 1

                # First check for implicit null values
                if self.prior_token in ('=', '%', ','):
                    if (self.token in (',', '/', '&', '$') and
                            not (self.prior_token == ',' and
                                 self.token in ('/', '&', '$'))):
                        self._append_value(v_values, None, v_idx, n_vals)

                elif self.prior_token == '*':
                    if self.token not in ('/', '&', '$'):
                        self._update_tokens()

                    if (self.prior_token == ',' or self.token == '='
                            or (self.token in ('/', '&', '$')
                                and self.prior_token == '*')):
                        next_value = None

                        # XXX: Repeated ,, after N*, will be off by one...
                        if self.prior_token == ',' and self.token == ',':
                            n_vals += 1

                    else:
                        next_value = self._parse_value()

                    self._append_value(v_values, next_value, v_idx, n_vals)

                else:
                    next_value = self._parse_value()
                    self._append_value(v_values, next_value, v_idx, n_vals)

                # Reset default repeat factor for subsequent values
                n_vals = 1

                # Exit for end of nml group (/, &, $) or null broadcast (=)
                if self.token in ('/', '&', '$', '='):
                    break
                else:
                    # NOTE: it is probably very inefficient to keep re-creating
                    # iterators upon every element; this solution reflects the
                    # absence of mature lookahead in the script.
                    #
                    # This is a temporary fix to address errors caused by
                    # patches of different length from the original value, and
                    # represents a direction to fully rewrite the parser using
                    # `tee`.

                    # NOTE: We may be able to assume that self.token is a value
                    #       rather than prepending it to the iterator.
                    self.tokens, pre_lookahead = itertools.tee(self.tokens)
                    lookahead = itertools.chain([self.token], pre_lookahead)
                    n_vals_remain = count_values(lookahead)

                    if patch_values:
                        # TODO: Patch indices that are not set in the namelist
                        if (p_idx < len(patch_values)
                                and n_vals_remain > 0
                                and self.token != ','):
                            p_val = patch_values[p_idx]
                            p_repr = patch_nml._f90repr(patch_values[p_idx])
                            p_idx += 1
                            self._update_tokens(override=p_repr)
                            if isinstance(p_val, complex):
                                # Skip over the complex content
                                # NOTE: Assumes input and patch are complex
                                self._update_tokens(write_token=False)
                                self._update_tokens(write_token=False)
                                self._update_tokens(write_token=False)
                                self._update_tokens(write_token=False)
                        else:
                            # Skip any values beyond the patch size
                            skip = (p_idx >= len(patch_values))
                            self._update_tokens(patch_skip=skip)
                    else:
                        self._update_tokens()

        if patch_values:
            v_values = patch_values

        if not v_idx:
            v_values = delist(v_values)

        return v_name, v_values

    def _parse_indices(self):
        """Parse a sequence of Fortran vector indices as a list of tuples."""
        v_name = self.prior_token
        v_indices = []

        while self.token in (',', '('):
            v_indices.append(self._parse_index(v_name))

        return v_indices

    def _parse_index(self, v_name):
        """Parse Fortran vector indices into a tuple of Python indices."""
        i_start = i_end = i_stride = None

        # Start index
        self._update_tokens()
        try:
            i_start = int(self.token)
            self._update_tokens()
        except ValueError:
            if self.token in (',', ')'):
                raise ValueError('{0} index cannot be empty.'.format(v_name))
            elif not self.token == ':':
                raise

        # End index
        if self.token == ':':
            self._update_tokens()
            try:
                i_end = 1 + int(self.token)
                self._update_tokens()
            except ValueError:
                if self.token == ':':
                    raise ValueError('{0} end index cannot be implicit '
                                     'when using stride.'.format(v_name))
                elif self.token not in (',', ')'):
                    raise
        elif self.token in (',', ')'):
            # Replace index with single-index range
            if i_start:
                i_end = 1 + i_start

        # Stride index
        if self.token == ':':
            self._update_tokens()
            try:
                i_stride = int(self.token)
            except ValueError:
                if self.token == ')':
                    raise ValueError('{0} stride index cannot be '
                                     'implicit.'.format(v_name))
                else:
                    raise

            if i_stride == 0:
                raise ValueError('{0} stride index cannot be zero.'
                                 ''.format(v_name))

            self._update_tokens()

        if self.token not in (',', ')'):
            raise ValueError('{0} index did not terminate '
                             'correctly.'.format(v_name))

        idx_triplet = (i_start, i_end, i_stride)
        return idx_triplet

    def _parse_value(self, write_token=True, override=None):
        """Convert string repr of Fortran type to equivalent Python type."""
        v_str = self.prior_token

        # Construct the complex string
        if v_str == '(':
            v_re = self.token

            self._update_tokens(write_token)
            assert self.token == ','

            self._update_tokens(write_token)
            v_im = self.token

            self._update_tokens(write_token)
            assert self.token == ')'

            self._update_tokens(write_token, override)
            v_str = '({0}, {1})'.format(v_re, v_im)

        recast_funcs = [int, pyfloat, pycomplex, pybool, pystr]

        for f90type in recast_funcs:
            try:
                # Unclever hack.. integrate this better
                if f90type == pybool:
                    value = pybool(v_str, self.strict_logical)
                else:
                    value = f90type(v_str)
                return value
            except ValueError:
                continue

    def _update_tokens(self, write_token=True, override=None,
                       patch_skip=False):
        """Update tokens to the next available values."""
        next_token = next(self.tokens)

        patch_value = ''
        patch_tokens = ''

        if self.pfile and write_token:
            token = override if override else self.token
            patch_value += token

        while next_token[0] in self.comment_tokens + whitespace:
            if self.pfile:
                if next_token[0] in self.comment_tokens:
                    while not next_token == '\n':
                        patch_tokens += next_token
                        next_token = next(self.tokens)
                patch_tokens += next_token

            # Several sections rely on StopIteration to terminate token search
            # If that occurs, dump the patched tokens immediately
            try:
                next_token = next(self.tokens)
            except StopIteration:
                if not patch_skip or next_token in ('=', '(', '%'):
                    patch_tokens = patch_value + patch_tokens

                if self.pfile:
                    self.pfile.write(patch_tokens)
                raise

        # Write patched values and whitespace + comments to file
        if not patch_skip or next_token in ('=', '(', '%'):
            patch_tokens = patch_value + patch_tokens

        if self.pfile:
            self.pfile.write(patch_tokens)

        # Update tokens, ignoring padding
        self.token, self.prior_token = next_token, self.token

    def _append_value(self, v_values, next_value, v_idx=None, n_vals=1):
        """Update a list of parsed values with a new value."""
        for _ in range(n_vals):
            if v_idx:
                try:
                    v_i = next(v_idx)
                except StopIteration:
                    # Repeating commas are null-statements and can be ignored
                    # Otherwise, we warn the user that this is a bad namelist
                    if next_value is not None:
                        warnings.warn(
                            'f90nml: warning: Value {v} is not assigned to '
                            'any variable and has been removed.'
                            ''.format(v=next_value)
                        )

                    # There are more values than indices, so we stop here
                    break

                v_s = [self.default_start_index if idx is None else idx
                       for idx in v_idx.first]

                if not self.row_major:
                    v_i = v_i[::-1]
                    v_s = v_s[::-1]

                # Multidimensional arrays
                if not self.sparse_arrays:
                    pad_array(v_values, list(zip(v_i, v_s)))

                # We iterate inside the v_values and inspect successively
                # deeper lists within the list tree.  If the requested index is
                # missing, we re-size that particular entry.
                # (NOTE: This is unnecessary when sparse_arrays is disabled.)

                v_subval = v_values
                for (i_v, i_s) in zip(v_i[:-1], v_s[:-1]):
                    try:
                        v_subval = v_subval[i_v - i_s]
                    except IndexError:
                        size = len(v_subval)
                        v_subval.extend([] for _ in range(size, i_v - i_s + 1))
                        v_subval = v_subval[i_v - i_s]

                # On the deepest level, we explicitly assign the value
                i_v, i_s = v_i[-1], v_s[-1]
                try:
                    v_subval[i_v - i_s] = next_value
                except IndexError:
                    size = len(v_subval)
                    v_subval.extend(None for _ in range(size, i_v - i_s + 1))
                    v_subval[i_v - i_s] = next_value
            else:
                v_values.append(next_value)


# Support functions
def prepad_array(var, v_start_idx, new_start_idx):
    """Return a resized vector based on the new start index."""
    prior_var = var[:]

    # Read the outer values
    i_p = v_start_idx[-1]
    i_v = new_start_idx[-1]

    # Compute the outer index padding
    if i_p is not None and i_v is not None and i_v < i_p:
        pad = [None for _ in range(i_p - i_v)]
    else:
        pad = []

    # Apply prepad rules to interior arrays
    for i, v in enumerate(var):
        if isinstance(v, list):
            prior_var[i] = prepad_array(v, v_start_idx[:-1],
                                        new_start_idx[:-1])
    return pad + prior_var


def pad_array(v, idx):
    """Expand lists in multidimensional arrays to pad unset values."""
    i_v, i_s = idx[0]

    if len(idx) > 1:
        # Append missing subarrays
        v.extend([[] for _ in range(len(v), i_v - i_s + 1)])

        # Pad elements
        for e in v:
            pad_array(e, idx[1:])
    else:
        v.extend([None for _ in range(len(v), i_v - i_s + 1)])


def merge_values(src, new):
    """Merge two lists or dicts into a single element."""
    if isinstance(src, dict) and isinstance(new, dict):
        return merge_dicts(src, new)
    else:
        if not isinstance(src, list):
            src = [src]
        if not isinstance(new, list):
            new = [new]

        return merge_lists(src, new)


def merge_lists(src, new):
    """Update a value list with a list of new or updated values."""
    l_min, l_max = (src, new) if len(src) < len(new) else (new, src)

    l_min.extend(None for i in range(len(l_min), len(l_max)))

    for i, val in enumerate(new):
        if isinstance(val, dict) and isinstance(src[i], dict):
            new[i] = merge_dicts(src[i], val)
        elif isinstance(val, list) and isinstance(src[i], list):
            new[i] = merge_lists(src[i], val)
        elif val is not None:
            new[i] = val
        else:
            new[i] = src[i]

    return new


def merge_dicts(src, patch):
    """Merge contents of dict `patch` into `src`."""
    for key in patch:
        if key in src:
            if isinstance(src[key], dict) and isinstance(patch[key], dict):
                merge_dicts(src[key], patch[key])
            else:
                src[key] = merge_values(src[key], patch[key])
        else:
            src[key] = patch[key]

    return src


def delist(values):
    """Reduce lists of zero or one elements to individual values."""
    assert isinstance(values, list)

    if not values:
        return None
    elif len(values) == 1:
        return values[0]

    return values


def count_values(tokens):
    """Identify the number of values ahead of the current token."""
    ntoks = 0
    for tok in tokens:
        if tok.isspace() or tok == ',':
            continue
        elif tok in ('=', '/', '$', '&'):
            if ntoks > 0 and tok == '=':
                ntoks -= 1
            break
        else:
            ntoks += 1

    return ntoks



"""Command line interface to f90nml.

:copyright: Copyright 2017 Marshall Ward, see AUTHORS for details.
:license: Apache License, Version 2.0, see LICENSE for details
"""
from __future__ import print_function

import warnings
import argparse
import json
import os
import sys
try:
    from StringIO import StringIO   # Python 2.x
except ImportError:
    from io import StringIO         # Python 3.x
try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict

import f90nml
try:
    import yaml
    has_yaml = True

    # Preserve ordering in YAML output
    #   https://stackoverflow.com/a/31609484/317172
    represent_dict_order = (lambda self, data:
                            self.represent_mapping('tag:yaml.org,2002:map',
                                                   data.items()))
    yaml.add_representer(OrderedDict, represent_dict_order)

except ImportError:
    has_yaml = False


def parse():
    """Parse the command line input arguments."""
    parser = argparse.ArgumentParser()

    parser.add_argument('--version', action='version',
                        version='f90nml {0}'.format(f90nml.__version__))

    parser.add_argument('--group', '-g', action='store',
                        help="specify namelist group to modify. "
                        "When absent, the first group is used")
    parser.add_argument('--variable', '-v', action='append',
                        help="specify the namelist variable to add or modify, "
                        "followed by the new value. Expressions are of the "
                        "form `VARIABLE=VALUE`")
    parser.add_argument('--patch', '-p', action='store_true',
                        help="modify the existing namelist as a patch")
    parser.add_argument('--format', '-f', action='store',
                        help="specify the output format (json, yaml, or nml)")

    parser.add_argument('input', nargs='?')
    parser.add_argument('output', nargs='?')

    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit()

    args = parser.parse_args()

    input_fname = args.input
    output_fname = args.output

    # Get input format
    # TODO: Combine with output format
    if input_fname:
        _, input_ext = os.path.splitext(input_fname)
        if input_ext == '.json':
            input_fmt = 'json'
        elif input_ext == '.yaml':
            input_fmt = 'yaml'
        else:
            input_fmt = 'nml'
    else:
        input_fmt = 'nml'

    # Output format flag validation
    valid_formats = ('json', 'yaml', 'nml')
    if args.format and args.format not in valid_formats:
        print('f90nml: error: format must be one of the following: {0}'
              ''.format(valid_formats), file=sys.stderr)
        sys.exit(-1)

    # Get output format
    # TODO: Combine with input format
    if not args.format:
        if output_fname:
            _, output_ext = os.path.splitext(output_fname)
            if output_ext == '.json':
                output_fmt = 'json'
            elif output_ext in ('.yaml', '.yml'):
                output_fmt = 'yaml'
            else:
                output_fmt = 'nml'
        else:
            output_fmt = 'nml'
    else:
        output_fmt = args.format

    # Confirm that YAML module is available
    if (input_fmt == 'yaml' or output_fmt == 'yaml') and not has_yaml:
        print('f90nml: error: YAML module could not be found.',
              file=sys.stderr)
        print('  To enable YAML support, install PyYAML or use the '
              'f90nml[yaml] package.', file=sys.stderr)
        sys.exit(-1)

    # Do not patch non-namelist output
    if any(fmt != 'nml' for fmt in (input_fmt, output_fmt)) and args.patch:
        print('f90nml: error: Only namelist files can be patched.',
              file=sys.stderr)
        sys.exit(-1)

    # Read the input file
    if input_fname:
        if input_fmt in ('json', 'yaml'):
            if input_fmt == 'json':
                with open(input_fname) as input_file:
                    input_data = json.load(input_file)
            elif input_ext == '.yaml':
                with open(input_fname) as input_file:
                    input_data = yaml.safe_load(input_file)
        else:
            input_data = f90nml.read(input_fname)
    else:
        input_data = {}

    input_data = f90nml.Namelist(input_data)

    # Construct the update namelist
    update_nml = {}
    if args.variable:
        if not args.group:
            # Use the first available group
            grp = list(input_data.keys())[0]
            warnings.warn(
                'f90nml: warning: Assuming variables are in group \'{g}\'.'
                ''.format(g=grp)
            )
        else:
            grp = args.group

        update_nml_str = '&{0} {1} /\n'.format(grp, ', '.join(args.variable))
        update_io = StringIO(update_nml_str)
        update_nml = f90nml.read(update_io)
        update_io.close()

    # Target output
    output_file = open(output_fname, 'w') if output_fname else sys.stdout

    if args.patch:
        # We have to read the file twice for a patch.  The main reason is
        # to identify the default group, in case this is not provided.
        # It could be avoided if a group is provided, but logically that could
        # a mess that I do not want to sort out right now.
        f90nml.patch(input_fname, update_nml, output_file)

    else:
        # Update the input namelist directly
        if update_nml:
            try:
                input_data[grp].update(update_nml[grp])
            except KeyError:
                input_data[grp] = update_nml[grp]

        # Write to output
        if not args.patch:
            if output_fmt in ('json', 'yaml'):
                if output_fmt == 'json':
                    input_data = input_data.todict(complex_tuple=True)
                    json.dump(input_data, output_file,
                              indent=4, separators=(',', ': '))
                    output_file.write('\n')

                elif output_fmt == 'yaml':
                    input_data = input_data.todict(complex_tuple=True)
                    yaml.dump(input_data, output_file,
                              default_flow_style=False)
            else:
                # Default to namelist output
                f90nml.write(input_data, output_file)

    # Cleanup
    if output_file != sys.stdout:
        output_file.close()



"""Conversion of Fortran values (as strings) to equivalent Python values.

The functions in this module are used to convert the string representation of
the values of basic Fortran data types into equivalent Python values.

:copyright: Copyright 2014 Marshall Ward, see AUTHORS for details.
:license: Apache License, Version 2.0, see LICENSE for details.
"""
import re


def pyfloat(v_str):
    """Convert string repr of Fortran floating point to Python double."""
    # NOTE: There is no loss of information from SP to DP floats

    return float(re.sub('(?<=[^eEdD])(?=[+-])', 'e',
                        v_str.lower().replace('d', 'e')))


def pycomplex(v_str):
    """Convert string repr of Fortran complex to Python complex."""
    assert isinstance(v_str, str)

    if v_str[0] == '(' and v_str[-1] == ')' and len(v_str.split(',')) == 2:
        v_re, v_im = v_str[1:-1].split(',', 1)

        # NOTE: Failed float(str) will raise ValueError
        return complex(pyfloat(v_re), pyfloat(v_im))
    else:
        raise ValueError('{0} must be in complex number form (x, y).'
                         ''.format(v_str))


def pybool(v_str, strict_logical=True):
    """Convert string repr of Fortran logical to Python logical."""
    assert isinstance(v_str, str)
    assert isinstance(strict_logical, bool)

    if strict_logical:
        v_bool = v_str.lower()
    else:
        try:
            if v_str.startswith('.'):
                v_bool = v_str[1].lower()
            else:
                v_bool = v_str[0].lower()
        except IndexError:
            raise ValueError('{0} is not a valid logical constant.'
                             ''.format(v_str))

    if v_bool in ('.true.', '.t.', 'true', 't'):
        return True
    elif v_bool in ('.false.', '.f.', 'false', 'f'):
        return False
    else:
        raise ValueError('{0} is not a valid logical constant.'.format(v_str))


def pystr(v_str):
    """Convert string repr of Fortran string to Python string."""
    assert isinstance(v_str, str)

    if v_str[0] in ("'", '"') and v_str[0] == v_str[-1]:
        quote = v_str[0]
        out = v_str[1:-1]
    else:
        # NOTE: This is non-standard Fortran.
        #       For example, gfortran rejects non-delimited strings.
        quote = None
        out = v_str

    # Replace escaped strings
    if quote:
        out = out.replace(2 * quote, quote)

    return out



"""Fortran namelist interface.

The ``Namelist`` is a representation of a Fortran namelist and its contents in
a Python environment.

:copyright: Copyright 2014 Marshall Ward, see AUTHORS for details.
:license: Apache License, Version 2.0, see LICENSE for details.
"""
from __future__ import print_function
import itertools
import copy
import numbers
import os
import platform
try:
    from StringIO import StringIO   # Python 2.x
except ImportError:
    from io import StringIO         # Python 3.x
try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict
try:
    basestring          # Python 2.x
except NameError:
    basestring = str    # Python 3.x


class Namelist(OrderedDict):
    """Representation of Fortran namelist in a Python environment.

    Namelists can be initialised as empty or with a pre-defined `dict` of
    `items`.  If an explicit default start index is required for `items`, then
    it can be initialised with the `default_start_index` input argument.

    In addition to the standard methods supported by `dict`, several additional
    methods and properties are provided for working with Fortran namelists.
    """

    class RepeatValue(object):
        """Container class for output using repeat counters."""

        def __init__(self, n, value):
            """Create the RepeatValue object."""
            self.repeats = n
            self.value = value

    def __init__(self, *args, **kwds):
        """Create the Namelist object."""
        s_args = list(args)

        # If using (unordered) dict, then resort the keys for reproducibility
        if (args and not isinstance(args[0], OrderedDict) and
                isinstance(args[0], dict)):
            s_args[0] = sorted(args[0].items())

        # Assign the default start index
        try:
            self._default_start_index = kwds.pop('default_start_index')
        except KeyError:
            self._default_start_index = None

        super(Namelist, self).__init__(*s_args, **kwds)

        self.start_index = self.pop('_start_index', {})

        # Update the complex tuples as intrinsics
        # TODO: We are effectively setting these twice.  Instead, fetch these
        # from s_args rather than relying on Namelist to handle the content.
        if '_complex' in self:
            for key in self['_complex']:
                if all(isinstance(v, list) for v in self[key]):
                    self[key] = [complex(*v) for v in self[key]]
                else:
                    self[key] = complex(*self[key])
            self.pop('_complex')

        # Formatting properties
        self._column_width = 72
        self._indent = 4 * ' '
        self._end_comma = False
        self._uppercase = False
        self._float_format = ''
        self._logical_repr = {False: '.false.', True: '.true.'}
        self._index_spacing = False
        self._repeat_counter = False
        self._split_strings = False

        # Namelist group spacing flag
        self._newline = False

        # Check for pre-set indentation
        self.indent = self.pop('_indent', self.indent)

        # PyPy 2 is dumb and does not use __setitem__() inside __init__()
        # This loop will explicitly convert any internal dicts to Namelists.
        if (platform.python_implementation() == 'PyPy' and
                platform.python_version_tuple()[0] == '2'):
            for key, value in self.items():
                self[key] = value

    def __contains__(self, key):
        """Case-insensitive interface to OrderedDict."""
        return super(Namelist, self).__contains__(key.lower())

    def __delitem__(self, key):
        """Case-insensitive interface to OrderedDict."""
        return super(Namelist, self).__delitem__(key.lower())

    def __getitem__(self, key):
        """Case-insensitive interface to OrderedDict."""
        if isinstance(key, basestring):
            return super(Namelist, self).__getitem__(key.lower())
        else:
            keyiter = iter(key)
            grp, var = next(keyiter).lower(), next(keyiter).lower()
            return super(Namelist, self).__getitem__(grp).__getitem__(var)

    def __setitem__(self, key, value):
        """Case-insensitive interface to OrderedDict.

        Python dict inputs to the Namelist, such as derived types, are also
        converted into Namelists.
        """
        if isinstance(value, dict) and not isinstance(value, Namelist):
            value = Namelist(value,
                             default_start_index=self.default_start_index)

        elif is_nullable_list(value, dict):
            for i, v in enumerate(value):
                if isinstance(v, Namelist) or v is None:
                    value[i] = v
                else:
                    # value is a non-Namelist dict
                    value[i] = Namelist(
                        v,
                        default_start_index=self.default_start_index
                    )

        super(Namelist, self).__setitem__(key.lower(), value)

    def __str__(self):
        """Print the Fortran representation of the namelist.

        Currently this can only be applied to the full contents of the namelist
        file.  Indiviual namelist groups or values may not render correctly.
        """
        output = StringIO()
        if all(isinstance(v, Namelist) for v in self.values()):
            self._writestream(output)
        else:
            print(repr(self), file=output)

        nml_string = output.getvalue().rstrip()
        output.close()
        return nml_string

    # Format configuration

    @property
    def column_width(self):
        """Set the maximum number of characters per line of the namelist file.

        :type: ``int``
        :default: 72

        Tokens longer than ``column_width`` are allowed to extend past this
        limit.
        """
        return self._column_width

    @column_width.setter
    def column_width(self, width):
        """Validate and set the column width."""
        if isinstance(width, int):
            if width >= 0:
                self._column_width = width
            else:
                raise ValueError('Column width must be nonnegative.')
        else:
            raise TypeError('Column width must be a nonnegative integer.')

    @property
    def default_start_index(self):
        """Set the default start index for vectors with no explicit index.

        :type: ``int``, ``None``
        :default: ``None``

        When the `default_start_index` is set, all vectors without an explicit
        start index are assumed to begin with `default_start_index`.  This
        index is shown when printing the namelist output.

        If set to `None`, then no start index is assumed and is left as
        implicit for any vectors undefined in `start_index`.
        """
        return self._default_start_index

    @default_start_index.setter
    def default_start_index(self, value):
        if not isinstance(value, int):
            raise TypeError('default_start_index must be an integer.')
        self._default_start_index = value

    @property
    def end_comma(self):
        """Append commas to the end of namelist variable entries.

        :type: ``bool``
        :default: ``False``

        Fortran will generally disregard any commas separating variable
        assignments, and the default behaviour is to omit these commas from the
        output.  Enabling this flag will append commas at the end of the line
        for each variable assignment.
        """
        return self._end_comma

    @end_comma.setter
    def end_comma(self, value):
        """Validate and set the comma termination flag."""
        if not isinstance(value, bool):
            raise TypeError('end_comma attribute must be a logical type.')
        self._end_comma = value

    @property
    def false_repr(self):
        """Set the string representation of logical false values.

        :type: ``str``
        :default: ``'.false.'``

        This is equivalent to the first element of ``logical_repr``.
        """
        return self._logical_repr[0]

    @false_repr.setter
    def false_repr(self, value):
        """Validate and set the logical false representation."""
        if isinstance(value, str):
            if not (value.lower().startswith('f') or
                    value.lower().startswith('.f')):
                raise ValueError("Logical false representation must start "
                                 "with 'F' or '.F'.")
            else:
                self._logical_repr[0] = value
        else:
            raise TypeError('Logical false representation must be a string.')

    @property
    def float_format(self):
        """Set the namelist floating point format.

        :type: ``str``
        :default: ``''``

        The property sets the format string for floating point numbers,
        following the format expected by the Python ``format()`` function.
        """
        return self._float_format

    @float_format.setter
    def float_format(self, value):
        """Validate and set the upper case flag."""
        if isinstance(value, str):
            # Duck-test the format string; raise ValueError on fail
            '{0:{1}}'.format(1.23, value)

            self._float_format = value
        else:
            raise TypeError('Floating point format code must be a string.')

    @property
    def indent(self):
        r"""Set the whitespace indentation of namelist entries.

        :type: ``int``, ``str``
        :default: ``'    '`` (four spaces)

        This can be set to an integer, denoting the number of spaces, or to an
        explicit whitespace character, such as a tab (``\t``).
        """
        return self._indent

    @indent.setter
    def indent(self, value):
        """Validate and set the indent width."""
        # Explicit indent setting
        if isinstance(value, str):
            if value.isspace() or len(value) == 0:
                self._indent = value
            else:
                raise ValueError('String indentation can only contain '
                                 'whitespace.')

        # Set indent width
        elif isinstance(value, int):
            if value >= 0:
                self._indent = value * ' '
            else:
                raise ValueError('Indentation spacing must be nonnegative.')

        else:
            raise TypeError('Indentation must be specified by string or space '
                            'width.')

    @property
    def index_spacing(self):
        """Apply a space between indexes of multidimensional vectors.

        :type: ``bool``
        :default: ``False``
        """
        return self._index_spacing

    @index_spacing.setter
    def index_spacing(self, value):
        """Validate and set the index_spacing flag."""
        if not isinstance(value, bool):
            raise TypeError('index_spacing attribute must be a logical type.')
        self._index_spacing = value

    # NOTE: This presumes that bools and ints are identical as dict keys
    @property
    def logical_repr(self):
        """Set the string representation of logical values.

        :type: ``dict``
        :default: ``{False: '.false.', True: '.true.'}``

        There are multiple valid representations of True and False values in
        Fortran.  This property sets the preferred representation in the
        namelist output.

        The properties ``true_repr`` and ``false_repr`` are also provided as
        interfaces to the elements of ``logical_repr``.
        """
        return self._logical_repr

    @logical_repr.setter
    def logical_repr(self, value):
        """Set the string representation of logical values."""
        if not any(isinstance(value, t) for t in (list, tuple)):
            raise TypeError("Logical representation must be a tuple with "
                            "a valid true and false value.")
        if not len(value) == 2:
            raise ValueError("List must contain two values.")

        self.false_repr = value[0]
        self.true_repr = value[1]

    @property
    def repeat_counter(self):
        """Return whether the namelist uses repeat counters for arrays.

        If True, then arrays with repeated values will use repeat tokens.  For
        example, the array ``[1, 2, 2, 2]`` will be written as ``1, 3*2``.

        :type: ``bool``
        :default: ``False``
        """
        return self._repeat_counter

    @repeat_counter.setter
    def repeat_counter(self, value):
        """Set whether array output should be done in repeat form."""
        if isinstance(value, bool):
            self._repeat_counter = value
        else:
            raise TypeError(r"repeat must be of type ``bool``")

    @property
    def split_strings(self):
        """Split strings at the ``column_width`` over multiple lines.

        :type: ``bool``
        :default: ``False``
        """
        return self._split_strings

    @split_strings.setter
    def split_strings(self, value):
        """Validate and set the split_strings flag."""
        if not isinstance(value, bool):
            raise TypeError('split_strings attribute must be a logical type.')
        self._split_strings = value

    @property
    def start_index(self):
        """Set the starting index for each vector in the namelist.

        :type: ``dict``
        :default: ``{}``

        ``start_index`` is stored as a dict which contains the starting index
        for each vector saved in the namelist.  For the namelist ``vec.nml``
        shown below,

        .. code-block:: fortran

           &vec_nml
               a = 1, 2, 3
               b(0:2) = 0, 1, 2
               c(3:5) = 3, 4, 5
               d(:,:) = 1, 2, 3, 4
           /

        the ``start_index`` contents are

        .. code:: python

           >>> import f90nml
           >>> nml = f90nml.read('vec.nml')
           >>> nml['vec_nml'].start_index
           {'b': [0], 'c': [3], 'd': [None, None]}

        The starting index of ``a`` is absent from ``start_index``, since its
        starting index is unknown and its values cannot be assigned without
        referring to the corresponding Fortran source.
        """
        return self._start_index

    @start_index.setter
    def start_index(self, value):
        """Validate and set the vector start index."""
        # TODO: Validate contents?  (May want to set before adding the data.)
        if not isinstance(value, dict):
            raise TypeError('start_index attribute must be a dict.')
        self._start_index = value

    @property
    def true_repr(self):
        """Set the string representation of logical true values.

        :type: ``str``
        :default: ``.true.``

        This is equivalent to the second element of ``logical_repr``.
        """
        return self._logical_repr[1]

    @true_repr.setter
    def true_repr(self, value):
        """Validate and set the logical true representation."""
        if isinstance(value, str):
            if not (value.lower().startswith('t') or
                    value.lower().startswith('.t')):
                raise ValueError("Logical true representation must start with "
                                 "'T' or '.T'.")
            else:
                self._logical_repr[1] = value
        else:
            raise TypeError('Logical true representation must be a string.')

    @property
    def uppercase(self):
        """Print group and variable names in uppercase.

        :type: ``bool``
        :default: ``False``

        This is equivalent to the second element of ``logical_repr``.
        """
        return self._uppercase

    @uppercase.setter
    def uppercase(self, value):
        """Validate and set the uppercase flag."""
        if not isinstance(value, bool):
            raise TypeError('uppercase attribute must be a logical type.')
        self._uppercase = value

    def write(self, nml_path, force=False, sort=False):
        """Write Namelist to a Fortran 90 namelist file.

        >>> nml = f90nml.read('input.nml')
        >>> nml.write('out.nml')
        """
        nml_is_file = hasattr(nml_path, 'read')
        if not force and not nml_is_file and os.path.isfile(nml_path):
            raise IOError('File {0} already exists.'.format(nml_path))

        nml_file = nml_path if nml_is_file else open(nml_path, 'w')
        try:
            self._writestream(nml_file, sort)
        finally:
            if not nml_is_file:
                nml_file.close()

    def patch(self, nml_patch):
        """Update the namelist from another partial or full namelist.

        This is different from the intrinsic `update()` method, which replaces
        a namelist section.  Rather, it updates the values within a section.
        """
        for sec in nml_patch:
            if sec not in self:
                self[sec] = Namelist()
            self[sec].update(nml_patch[sec])

    def groups(self):
        """Return an iterator that spans values with group and variable names.

        Elements of the iterator consist of a tuple containing two values.  The
        first is internal tuple containing the current namelist group and its
        variable name.  The second element of the returned tuple is the value
        associated with the current group and variable.
        """
        for key, value in self.items():
            for inner_key, inner_value in value.items():
                yield (key, inner_key), inner_value

    def _writestream(self, nml_file, sort=False):
        """Output Namelist to a streamable file object."""
        # Reset newline flag
        self._newline = False

        if sort:
            sel = Namelist(sorted(self.items(), key=lambda t: t[0]))
        else:
            sel = self

        for grp_name, grp_vars in sel.items():
            # Check for repeated namelist records (saved as lists)
            if isinstance(grp_vars, list):
                for g_vars in grp_vars:
                    self._write_nmlgrp(grp_name, g_vars, nml_file, sort)
            else:
                self._write_nmlgrp(grp_name, grp_vars, nml_file, sort)

    def _write_nmlgrp(self, grp_name, grp_vars, nml_file, sort=False):
        """Write namelist group to target file."""
        if self._newline:
            print(file=nml_file)
        self._newline = True

        if self.uppercase:
            grp_name = grp_name.upper()

        if sort:
            grp_vars = Namelist(sorted(grp_vars.items(), key=lambda t: t[0]))

        print('&{0}'.format(grp_name), file=nml_file)

        for v_name, v_val in grp_vars.items():

            v_start = grp_vars.start_index.get(v_name, None)

            for v_str in self._var_strings(v_name, v_val, v_start=v_start):
                print(v_str, file=nml_file)

        print('/', file=nml_file)

    def _var_strings(self, v_name, v_values, v_idx=None, v_start=None):
        """Convert namelist variable to list of fixed-width strings."""
        if self.uppercase:
            v_name = v_name.upper()

        var_strs = []

        # Parse a multidimensional array
        if is_nullable_list(v_values, list):
            if not v_idx:
                v_idx = []

            i_s = v_start[::-1][len(v_idx)] if v_start else None

            # FIXME: We incorrectly assume 1-based indexing if it is
            # unspecified.  This is necessary because our output method always
            # separates the outer axes to one per line.  But we cannot do this
            # if we don't know the first index (which we are no longer assuming
            # to be 1-based elsewhere).  Unfortunately, the solution needs a
            # rethink of multidimensional output.

            # NOTE: Fixing this would also clean up the output of todict(),
            # which is now incorrectly documenting unspecified indices as 1.

            # For now, we will assume 1-based indexing here, just to keep
            # things working smoothly.
            if i_s is None:
                i_s = 1

            for idx, val in enumerate(v_values, start=i_s):
                v_idx_new = v_idx + [idx]
                v_strs = self._var_strings(v_name, val, v_idx=v_idx_new,
                                           v_start=v_start)
                var_strs.extend(v_strs)

        # Parse derived type contents
        elif isinstance(v_values, Namelist):
            for f_name, f_vals in v_values.items():
                v_title = '%'.join([v_name, f_name])

                v_start_new = v_values.start_index.get(f_name, None)

                v_strs = self._var_strings(v_title, f_vals,
                                           v_start=v_start_new)
                var_strs.extend(v_strs)

        # Parse an array of derived types
        elif is_nullable_list(v_values, Namelist):
            if not v_idx:
                v_idx = []

            i_s = v_start[::-1][len(v_idx)] if v_start else 1

            for idx, val in enumerate(v_values, start=i_s):

                # Skip any empty elements in a list of derived types
                if val is None:
                    continue

                v_title = v_name + '({0})'.format(idx)

                v_strs = self._var_strings(v_title, val)
                var_strs.extend(v_strs)

        else:
            use_default_start_index = False
            if not isinstance(v_values, list):
                v_values = [v_values]
                use_default_start_index = False
            else:
                use_default_start_index = self.default_start_index is not None

            # Print the index range

            # TODO: Include a check for len(v_values) to determine if vector
            if v_idx or v_start or use_default_start_index:
                v_idx_repr = '('

                if v_start or use_default_start_index:
                    if v_start:
                        i_s = v_start[0]
                    else:
                        i_s = self.default_start_index

                    if i_s is None:
                        v_idx_repr += ':'

                    else:
                        i_e = i_s + len(v_values) - 1

                        if i_s == i_e:
                            v_idx_repr += '{0}'.format(i_s)
                        else:
                            v_idx_repr += '{0}:{1}'.format(i_s, i_e)
                else:
                    v_idx_repr += ':'

                if v_idx:
                    idx_delim = ', ' if self._index_spacing else ','
                    v_idx_repr += idx_delim
                    v_idx_repr += idx_delim.join(str(i) for i in v_idx[::-1])

                v_idx_repr += ')'

            else:
                v_idx_repr = ''

            # Split output across multiple lines (if necessary)
            v_header = self.indent + v_name + v_idx_repr + ' = '
            val_strs = []
            val_line = v_header

            if self._repeat_counter:
                v_values = list(
                    self.RepeatValue(len(list(x)), val)
                    for val, x in itertools.groupby(v_values)
                )

            for i_val, v_val in enumerate(v_values):
                # Increase column width if the header exceeds this value
                if len(v_header) >= self.column_width:
                    column_width = len(v_header) + 1
                else:
                    column_width = self.column_width

                if len(val_line) < column_width:
                    # NOTE: We allow non-strings to extend past the column
                    #   limit, but strings will be split as needed.
                    v_str = self._f90repr(v_val)

                    # Set a comma placeholder if needed
                    if i_val < len(v_values) - 1 or self.end_comma:
                        v_comma = ', '
                    else:
                        v_comma = ''

                    if self.split_strings and isinstance(v_val, str):
                        idx = column_width - len(val_line + v_comma.rstrip())

                        # Split the line along idx until we either exceed the
                        #   column width, or read the end of the string.
                        v_l, v_r = v_str[:idx], v_str[idx:]

                        if v_r:
                            # Check if string can fit on the next line
                            new_val_line = (
                                ' ' * len(v_header) + v_str + v_comma
                            )
                            if len(new_val_line.rstrip()) <= column_width:
                                val_strs.append(val_line)
                                val_line = ' ' * len(v_header)
                            else:
                                # Split string across multiple lines
                                while v_r:
                                    val_line += v_l
                                    val_strs.append(val_line)
                                    val_line = ''

                                    idx = column_width - len(v_comma.rstrip())
                                    v_l, v_r = v_r[:idx], v_r[idx:]

                                v_str = v_l

                    val_line += v_str + v_comma

                # Line break
                if len(val_line) >= column_width:
                    # Append current line to list of lines
                    val_strs.append(val_line.rstrip())

                    # Start new line with space corresponding to header
                    val_line = ' ' * len(v_header)

            # Append any remaining values
            if val_line and not val_line.isspace():
                val_strs.append(val_line.rstrip())

            # Final null values must always precede a comma
            if val_strs and v_values[-1] is None:
                # NOTE: val_strs has been rstrip-ed so lead with a space
                val_strs[-1] += ' ,'

            # Complete the set of values
            if val_strs:
                var_strs.extend(val_strs)

        return var_strs

    def todict(self, complex_tuple=False):
        """Return a dict equivalent to the namelist.

        Since Fortran variables and names cannot start with the ``_``
        character, any keys starting with this token denote metadata, such as
        starting index.

        The ``complex_tuple`` flag is used to convert complex data into an
        equivalent 2-tuple, with metadata stored to flag the variable as
        complex.  This is primarily used to facilitate the storage of the
        namelist into an equivalent format which does not support complex
        numbers, such as JSON or YAML.
        """
        # TODO: Preserve ordering
        nmldict = OrderedDict(self)

        # Search for namelists within the namelist
        # TODO: Move repeated stuff to new functions
        for key, value in self.items():
            if isinstance(value, Namelist):
                nml = copy.deepcopy(value)
                nmldict[key] = nml.todict(complex_tuple)

            elif isinstance(value, complex) and complex_tuple:
                nmldict[key] = [value.real, value.imag]
                try:
                    nmldict['_complex'].append(key)
                except KeyError:
                    nmldict['_complex'] = [key]

            elif isinstance(value, list):
                complex_list = False
                for idx, entry in enumerate(value):
                    if isinstance(entry, Namelist):
                        nml = copy.deepcopy(entry)
                        nmldict[key][idx] = nml.todict(complex_tuple)

                    elif isinstance(entry, complex) and complex_tuple:
                        nmldict[key][idx] = [entry.real, entry.imag]
                        complex_list = True

                if complex_list:
                    try:
                        nmldict['_complex'].append(key)
                    except KeyError:
                        nmldict['_complex'] = [key]

        # Append the start index if present
        if self.start_index:
            nmldict['_start_index'] = self.start_index

        return nmldict

    def _f90repr(self, value):
        """Convert primitive Python types to equivalent Fortran strings."""
        if isinstance(value, self.RepeatValue):
            return self._f90repeat(value)
        elif isinstance(value, bool):
            return self._f90bool(value)
        elif isinstance(value, numbers.Integral):
            return self._f90int(value)
        elif isinstance(value, numbers.Real):
            return self._f90float(value)
        elif isinstance(value, numbers.Complex):
            return self._f90complex(value)
        elif isinstance(value, basestring):
            return self._f90str(value)
        elif value is None:
            return ''
        else:
            raise ValueError('Type {0} of {1} cannot be converted to a Fortran'
                             ' type.'.format(type(value), value))

    def _f90repeat(self, value):
        """Return a Fortran 90 representation of a repeated value."""
        if value.repeats == 1:
            return self._f90repr(value.value)
        else:
            return "{0}*{1}".format(value.repeats,
                                    self._f90repr(value.value))

    def _f90bool(self, value):
        """Return a Fortran 90 representation of a logical value."""
        return self.logical_repr[value]

    def _f90int(self, value):
        """Return a Fortran 90 representation of an integer."""
        return str(value)

    def _f90float(self, value):
        """Return a Fortran 90 representation of a floating point number."""
        return '{0:{fmt}}'.format(value, fmt=self.float_format)

    def _f90complex(self, value):
        """Return a Fortran 90 representation of a complex number."""
        return '({0:{fmt}}, {1:{fmt}})'.format(value.real, value.imag,
                                               fmt=self.float_format)

    def _f90str(self, value):
        """Return a Fortran 90 representation of a string."""
        # Replace Python quote escape sequence with Fortran
        result = repr(str(value)).replace("\\'", "''").replace('\\"', '""')

        # Un-escape the Python backslash escape sequence
        result = result.replace('\\\\', '\\')

        return result


def is_nullable_list(val, vtype):
    """Return True if list contains either values of type `vtype` or None."""
    return (isinstance(val, list) and
            any(isinstance(v, vtype) for v in val) and
            all((isinstance(v, vtype) or v is None) for v in val))



"""Fortran namelist tokenizer.

:copyright: Copyright 2017 Marshall Ward, see AUTHORS for details.
:license: Apache License, Version 2.0, see LICENSE for details.
"""
import string


class Tokenizer(object):
    """Fortran namelist tokenizer."""

    # I don't use these two
    special_chars = ' =+-*/\\()[]{},.:;!"%&~<>?\'`|$#@'     # Table 3.1
    lexical_tokens = '=+-*/()[],.:;%&<>'                    # Meaningful?

    # I only use this one
    punctuation = '=*/\\()[]{},:;%&~<>?`|$#@'    # Unhandled Table 3.1 tokens

    def __init__(self):
        """Initialise the tokenizer."""
        self.characters = None
        self.prior_char = None
        self.char = None
        self.idx = None
        self.whitespace = string.whitespace.replace('\n', '')
        self.prior_delim = None

        # Set to true if inside a namelist group
        self.group_token = None

        # Standard token sets
        self.comment_tokens = '!'

    def parse(self, line):
        """Tokenize a line of Fortran source."""
        tokens = []

        self.idx = -1   # Bogus value to ensure idx = 0 after first iteration
        self.characters = iter(line)
        self.update_chars()

        while self.char != '\n':
            # Update namelist group status
            if self.char in ('&', '$'):
                self.group_token = self.char

            if self.group_token and (
                    (self.group_token, self.char) in (('&', '/'), ('$', '$'))):
                self.group_token = False

            word = ''
            if self.char in self.whitespace:
                while self.char in self.whitespace:
                    word += self.char
                    self.update_chars()

            elif self.char in self.comment_tokens or self.group_token is None:
                # Abort the iteration and build the comment token
                word = line[self.idx:-1]
                self.char = '\n'

            elif self.char in '"\'' or self.prior_delim:
                word = self.parse_string()

            elif self.char in Tokenizer.punctuation:
                word = self.char
                self.update_chars()

            else:
                while (not self.char.isspace()
                       and self.char not in Tokenizer.punctuation):
                    word += self.char
                    self.update_chars()

            tokens.append(word)

        return tokens

    def parse_string(self):
        """Tokenize a Fortran string."""
        word = ''

        if self.prior_delim:
            delim = self.prior_delim
            self.prior_delim = None
        else:
            delim = self.char
            word += self.char
            self.update_chars()

        while True:
            if self.char == delim:
                # Check for escaped delimiters
                self.update_chars()
                if self.char == delim:
                    word += 2 * delim
                    self.update_chars()
                else:
                    word += delim
                    break
            elif self.char == '\n':
                self.prior_delim = delim
                break
            else:
                word += self.char
                self.update_chars()

        return word

    def update_chars(self):
        """Update the current charters in the tokenizer."""
        # NOTE: We spoof non-Unix files by returning '\n' on StopIteration
        self.prior_char, self.char = self.char, next(self.characters, '\n')
        self.idx += 1



"""Installation script for f90nml.

Additional configuration settings are in ``setup.cfg``.

:copyright: Copyright 2014 Marshall Ward, see AUTHORS for details.
:license: Apache License, Version 2.0, see LICENSE for details.
"""
import os
import sys

try:
    from setuptools import setup
    from setuptools import Command
except ImportError:
    from distutils.core import setup
    from distutils.core import Command

import tests.test_f90nml

# Project details
project_name = 'f90nml'
project_version = __import__(project_name).__version__
project_readme_fname = 'README.rst'
project_scripts = [os.path.join('bin', f) for f in os.listdir('bin')]


# Test suite
class ProjectTest(Command):
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        unittest = tests.test_f90nml.unittest
        testcase = tests.test_f90nml.Test
        suite = unittest.TestLoader().loadTestsFromTestCase(testcase)
        result = unittest.TextTestRunner(verbosity=2).run(suite)
        sys.exit(not result.wasSuccessful())


# README
with open(project_readme_fname) as f:
    project_readme = f.read()


# Project setup
setup(
    name = project_name,
    version = project_version,
    description = 'Fortran 90 namelist parser',
    long_description = project_readme,
    author = 'Marshall Ward',
    author_email = 'f90nml@marshallward.org',
    url = 'http://github.com/marshallward/f90nml',

    packages = ['f90nml'],
    scripts=project_scripts,

    extras_require = {
        'yaml': ['PyYAML'],
    },

    cmdclass = {'test': ProjectTest},

    classifiers = [
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Utilities',
    ]
)



from __future__ import print_function

import math
import os
import sys
import unittest

try:
    from StringIO import StringIO   # Python 2.x
except ImportError:
    from io import StringIO         # Python 3.x

try:
    from collections import OrderedDict     # Python 3.x
except ImportError:
    from ordereddict import OrderedDict     # Python 2.x

try:
    import numpy
    has_numpy = True
except ImportError:
    has_numpy = False

try:
    import yaml
    has_yaml = True
except ImportError:
    has_yaml = False

# Include parent path in case we are running within the tests directory
sys.path.insert(1, '../')
import f90nml
import f90nml.cli
from f90nml.fpy import pybool
from f90nml.namelist import Namelist
from f90nml.findex import FIndex


class Test(unittest.TestCase):

    def setUp(self):
        # Move to test directory if running from setup.py
        if os.path.basename(os.getcwd()) != 'tests':
            os.chdir('tests')

        # Construct the reference namelist values

        self.empty_file = {}

        self.empty_nml = {'empty_nml': {}}

        self.null_nml = {
            'null_nml': {'null_value': None},
            'null_comma_nml': {'null_comma': None},
            'null_nocomma_rpt_nml': {
                'null_one': None,
                'null_two': None,
            }
        }

        self.unset_nml = {
            'unset_nml': {
                'x': None,
                'y': None
            }
        }

        self.types_nml = {
            'types_nml': {
                'v_integer': 1,
                'v_float': 1.0,
                'v_complex': 1+2j,
                'v_logical': True,
                'v_string': 'Hello',
            }
        }

        self.vector_nml = {
            'vector_nml': {
                'v': [1, 2, 3, 4, 5],
                'v_idx': [1, 2, 3, 4],
                'v_idx_ooo': [1, 2, 3, 4],
                'v_range': [1, 2, 3, 4],
                'v_start_zero': [1, 2, 3, 4],
                'v_start_minusone': [1, 2, 3, 4, 5],
                'v_zero_adj': [1, None, 3, 4],
                'v_zero_adj_ooo': [1, None, 3, 4],
                'v_implicit_start': [1, 2, 3, 4],
                'v_implicit_end': [1, 2, 3, 4],
                'v_implicit_all': [1, 2, 3, 4],
                'v_null_start': [None, 2, 3, 4],
                'v_null_interior': [1, 2, None, 4],
                'v_null_end': [1, 2, 3, None],
                'v_zero': [1, 0, 3],
                'v_stride': [1, None, 3, None, 5, None, 7],
                'v_single': [1],
                'v_implicit_merge': [1, 2],
                'v_explicit_merge': [1, 2],
                'v_complex': [1+2j, 3+4j, 5+6j],
            }
        }

        self.str_array_nml = {
            'str_array_nml': {
                'names': [
                    'abcdefg',
                    'tuvwxyz',
                    'abc123',
                    '123abc',
                    'abcdefg',
                    'abcdefghijklmnopqrstuvwxyz',
                    'xyz321',
                ], 
            }
        }

        self.multidim_nml = {
            'multidim_nml': {
                'v2d': [[1, 2], [3, 4]],
                'v3d': [[[1, 2], [3, 4]], [[5, 6], [7, 8]]],
                'w3d': [[[1, 2, 3, 4],
                         [5, 6, 7, 8],
                         [9, 10, 11, 12]],
                        [[13, 14, 15, 16],
                         [17, 18, 19, 20],
                         [21, 22, 23, 24]]],
                'v2d_explicit': [[1, 2], [3, 4]],
                'v2d_outer': [[1], [2], [3], [4]],
                'v2d_inner': [[1, 2, 3, 4]],
                'v2d_sparse': [[1, 2], [None, None], [5, 6]]
            }
        }

        self.multidim_ooo_nml = {
            'multidim_ooo_nml': {
                'a': [[1], [None, 2]],
                'b': [[1], [None, None, 3]],
            }
        }

        self.md_rowmaj_nml = {
            'multidim_nml': {
                'v2d': [[1, 3], [2, 4]],
                'v3d': [[[1, 5], [3, 7]], [[2, 6], [4, 8]]],
                'w3d': [[[1, 13], [5, 17], [9, 21]],
                        [[2, 14], [6, 18], [10, 22]],
                        [[3, 15], [7, 19], [11, 23]],
                        [[4, 16], [8, 20], [12, 24]]],
                'v2d_explicit': [[1, 3], [2, 4]],
                'v2d_outer': [[1, 2, 3, 4]],
                'v2d_inner': [[1], [2], [3], [4]],
                'v2d_sparse': [[1, None, 5], [2, None, 6]]
            }
        }

        self.dense_array_nml = {
            'sparse_array_nml': {
                'x': [
                    [1, None, None],
                    [None, None, None],
                    [None, None, 2],
                ]
            }
        }

        self.sparse_array_nml = {
            'sparse_array_nml': {
                'x': [
                    [1],
                    [],
                    [None, None, 2],
                ]
            }
        }

        self.default_one_index_nml = {
            'default_index_nml': {
                'v': [1, 2, 3, 4, 5]
            }
        }

        self.default_zero_index_nml = {
            'default_index_nml': {
                'v': [1, 2, None, 3, 4, 5]
            }
        }

        self.global_index_nml = {
            'global_index_nml': {
                'v_zero': [1, 2, 3, 4],
                'v_neg': [1, 2, 3, 4],
                'v_pos': [None, 1, 2, 3, 4]
            }
        }

        self.float_nml = {
            'float_nml': {
                'v_float': 1.,
                'v_decimal_start': .1,
                'v_decimal_end': 1.,
                'v_negative': -1.,
                'v_single': 1.,
                'v_double': 1.,
                'v_single_upper': 1.,
                'v_double_upper': 1.,
                'v_positive_index': 10.,
                'v_negative_index': 0.1,
                'v_no_exp_pos': 1.,
                'v_no_exp_neg': 1.,
                'v_no_exp_pos_dot': 1.,
                'v_no_exp_neg_dot': 1.,
                'v_neg_no_exp_pos': -1.,
                'v_neg_no_exp_neg': -1.,
                'v_pos_decimal': 0.01,
                'v_neg_decimal': -0.01,
            }
        }

        self.string_nml = {
            'string_nml': {
                'str_basic': 'hello',
                'str_no_delim': 'hello',
                'str_no_delim_token': '+hello',
                'str_no_delim_no_esc': "a''b",
                'single_esc_delim': "a 'single' delimiter",
                'double_esc_delim': 'a "double" delimiter',
                'double_nested': "''x'' \"y\"",
                'str_list': ['a', 'b', 'c'],
                'slist_no_space': ['a', 'b', 'c'],
                'slist_no_quote': ['a', 'b', 'c'],
                'slash': 'back\\slash',
            }
        }

        self.string_multiline_nml = {
            'string_multiline_nml': {
                'empty': '',
                'trailing_whitespace': '  '
            }
        }

        self.dtype_nml = {
            'dtype_nml': {
                'dt_scalar': {'val': 1},
                'dt_stack': {'outer': {'inner': 2}},
                'dt_vector': {'vec': [1, 2, 3]}
            },
            'dtype_multi_nml': {
                'dt': {
                    'x': 1,
                    'y': 2,
                    'z': 3,
                }
            },
            'dtype_nested_nml': {
                'f': {
                    'g': {
                        'x': 1,
                        'y': 2,
                        'z': 3,
                    }
                }
            },
            'dtype_field_idx_nml': {
                'f': {
                    'x': [1, 2, 3]}
                },
            'dtype_vec_nml': {
                'a': {
                    'b': [
                        {'c': 1, 'd': 2},
                        {'c': 3, 'd': 4},
                        {'c': 5, 'd': 6}
                    ]
                }
            },
            'dtype_sparse_vec_nml': {
                'a': {
                    'b': [{'c': 2}]     # NOTE: start_index is 2
                }
            },
            'dtype_single_value_vec_nml': {
                'a': [{'b': 1}]
            },
            'dtype_single_vec_merge_nml': {
                'a': {
                    'b': [{'c': 1, 'd': 2}]
                }
            },
            'dtype_list_ooo_noidx': {
                'a': [
                    {'i': 123},
                    {'i': 456},
                    {'i': 789},
                ],
            },
        }

        self.dtype_case_nml = {
            'dtype_mixed': {
                'b': {
                    'c_d_e': [{'id': 1}, {'id': 2}]
                }
            },
            'dtype_list_in_list': {
                'b': {
                    'c': [
                        {'id': 1},
                        {'id': 2},
                        {'id': 3},
                        {'id': 4, 'd': {'e': [10, 11]}}
                    ]
                }
            },
            'dtype_upper_scalar': {
                'b': {
                    'c': 1,
                    'd': [{'id': 2}],
                }
            },
            'dtype_upper_list': {
                'b': {
                    'c': [{'id': 1}, {'id': 2}]
                }
            },
            'dtype_index_overwrite': {
                'b': {
                    'c': [{'d': 1, 'e': 2, 'f': 3, 'g': 4, 'h': 5}]
                }
            },
            'dtype_list_staggered': {
                'b': {
                    'c': [
                        {'a': 1}, None, None, {'a': 1},
                        None, None, None, {'a': 1}
                    ]
                }
            },
        }

        self.bcast_nml = {
            'bcast_nml': {
                'x': [2.0, 2.0],
                'y': [None, None, None],
                'z': [True, True, True, True],
            },
            'bcast_nml_comma': {
                'x': [2.0, 2.0],
                'y': [None, None, None],
                'z': [True, True, True, True],
            },
            'bcast_endnull_nml': {
                'x': [2.0, 2.0],
                'y': [None, None, None],
            },
            'bcast_endnull_nml_comma': {
                'x': [2.0, 2.0],
                'y': [None, None, None],
            },
            'bcast_mixed_nml': {
                'x': [1, 1, 1, 2, 3, 4],
                'y': [1, 1, 1, 2, 2, 3],
            },
            'bcast_mixed_null_nml': {
                'x': [1, None, None, None, 3, 4],
                'y': [1, 1, 1, None, None, None, 3, 4],
                'z': [1, None, None, None, None, 4],
            },
        }

        self.comment_nml = {
            'comment_nml': {
                'v_cmt_inline': 123,
                'v_cmt_in_str': 'This token ! is not a comment',
                'v_cmt_after_str': 'This ! is not a comment',
            }
        }

        self.comment_alt_nml = {
            'comment_alt_nml': {
                'x': 1,
                'z': 3}
        }

        self.grp_repeat_nml = {
            'grp_repeat_nml': [{'x': 1}, {'x': 2}],
            'case_check_nml': [{'y': 1}, {'y': 2}],
        }

        self.f77_nml = {
            'f77_nml': {'x': 123},
            'next_f77_nml': {'y': 'abc'},
        }

        self.dollar_nml = {'dollar_nml': {'v': 1.}}

        self.multiline_nml = {
            'multiline_nml': {
                'x': [
                    1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
                    1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
                    1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
                    1, 1, 1, 1, 1, 1, 1, 1
                ]
            }
        }

        self.long_string_nml = {
            'long_string_nml': {
                's': ' '.join(12 * ['abcdefghijklmnopqrstuvwxyz'])
            }
        }

        self.ext_token_nml = {'ext_token_nml': {'x': 1}}

        self.list_patch_nml = {
            'list_patch_nml': {
                'x': ['1', '2', '3', '4', '5']
            }
        }

        self.repatch_nml = {
            'repatch_nml': {
                'x': [5, 6],
                'y': {'z': 7}
            },
            'newgrp_nml': {
                'a': 1,
                'b': 2,
            }
        }

        self.winfmt_nml = {'blah': {'blah': 1}}

        self.extern_cmt_nml = {
            'efitin': {
                'abc': 0,
            }
        }

        self.ieee_nml = {
            'ieee_nml': {
                'base_inf': float('inf'),
                'neg_inf': float('-inf'),
                'plus_inf': float('inf'),
                'base_nan': float('nan'),
                'plus_nan': float('nan'),
                'neg_nan': float('nan'),
            }
        }

        if has_numpy:
            self.numpy_nml = {
                'numpy_nml': OrderedDict((
                        ('np_integer', numpy.int64(1)),
                        ('np_float', numpy.float64(1.0)),
                        ('np_complex', numpy.complex128(1+2j)),
                    )
                )
            }

        if os.path.isfile('tmp.nml'):
            os.remove('tmp.nml')

    # Support functions
    def assert_file_equal(self, source_fname, target_fname):
        with open(source_fname) as source:
            with open(target_fname) as target:
                source_str = source.read()
                target_str = target.read()
                self.assertEqual(source_str, target_str)

    def assert_write(self, nml, target_fname, sort=False):
        self.assert_write_path(nml, target_fname, sort)
        self.assert_write_file(nml, target_fname, sort)

    def assert_write_path(self, nml, target_fname, sort=False):
        tmp_fname = 'tmp.nml'
        f90nml.write(nml, tmp_fname, sort=sort)
        try:
            self.assert_file_equal(tmp_fname, target_fname)
        finally:
            os.remove(tmp_fname)

    def assert_write_file(self, nml, target_fname, sort=False):
        tmp_fname = 'tmp.nml'
        with open(tmp_fname, 'w') as tmp_file:
            f90nml.write(nml, tmp_file, sort=sort)
            self.assertFalse(tmp_file.closed)
        try:
            self.assert_file_equal(tmp_fname, target_fname)
        finally:
            os.remove(tmp_fname)

    def get_cli_output(self, args, get_stderr=False):
        argv_in, stdout_in, stderr_in = sys.argv, sys.stdout, sys.stderr

        sys.argv = args
        sys.stdout = StringIO()
        sys.stderr = StringIO()

        try:
            f90nml.cli.parse()
        except SystemExit:
            pass

        sys.stdout.seek(0)
        stdout = sys.stdout.read()
        sys.stdout.close()

        sys.stderr.seek(0)
        stderr = sys.stderr.read()
        sys.stderr.close()

        sys.argv, sys.stdout, sys.stderr = argv_in, stdout_in, stderr_in

        # TODO: Some way to check both would be more useful...
        if get_stderr:
            return stderr
        else:
            return stdout

    # Tests
    def test_empty_file(self):
        test_nml = f90nml.read('empty_file')
        self.assertEqual(self.empty_file, test_nml)

    def test_empty_nml(self):
        test_nml = f90nml.read('empty.nml')
        self.assertEqual(self.empty_nml, test_nml)
        self.assert_write(test_nml, 'empty.nml')

    def test_null(self):
        test_nml = f90nml.read('null.nml')
        self.assertEqual(self.null_nml, test_nml)
        self.assert_write(test_nml, 'null_target.nml')

    def test_unset(self):
        test_nml = f90nml.read('unset.nml')
        self.assertEqual(self.unset_nml, test_nml)
        self.assert_write(test_nml, 'unset.nml')

    def test_types(self):
        test_nml = f90nml.read('types.nml')
        self.assertEqual(self.types_nml, test_nml)
        self.assert_write(test_nml, 'types.nml')

    def test_vector(self):
        test_nml = f90nml.read('vector.nml')
        self.assertEqual(self.vector_nml, test_nml)
        self.assert_write(test_nml, 'vector_target.nml')

    def test_vector_default_index(self):
        test_nml = f90nml.read('vector.nml')
        test_nml.default_start_index = 2
        self.assertEqual(self.vector_nml, test_nml)
        self.assert_write(test_nml, 'vector_default_idx.nml')

    def test_string_array(self):
        test_nml = f90nml.read('string_array.nml')
        test_nml.split_strings = True
        self.assertEqual(self.str_array_nml, test_nml)
        self.assert_write(test_nml, 'string_array_target.nml')

    def test_multidim(self):
        test_nml = f90nml.read('multidim.nml')
        self.assertEqual(self.multidim_nml, test_nml)
        self.assert_write(test_nml, 'multidim_target.nml')

    def test_multidim_ooo(self):
        test_nml = f90nml.read('multidim_ooo.nml')
        self.assertEqual(self.multidim_ooo_nml, test_nml)
        self.assert_write(test_nml, 'multidim_ooo_target.nml')

    def test_rowmaj_multidim(self):
        parser = f90nml.Parser()
        parser.row_major = True
        test_nml = parser.read('multidim.nml')
        self.assertEqual(self.md_rowmaj_nml, test_nml)

    def test_dense_arrays(self):
        parser = f90nml.Parser()
        test_nml = parser.read('sparse_array.nml')
        self.assertEqual(self.dense_array_nml, test_nml)

    def test_sparse_arrays(self):
        parser = f90nml.Parser()
        parser.sparse_arrays = True
        test_nml = parser.read('sparse_array.nml')
        self.assertEqual(self.sparse_array_nml, test_nml)

    def test_vector_too_long(self):
        sys_stderr = sys.stderr
        sys.stderr = StringIO()
        test_nml = f90nml.read('values_exceed_index.nml')
        # TODO: Check values
        sys.stderr = sys_stderr

    def test_parser_property_invalid(self):
        parser = f90nml.Parser()
        self.assertRaises(TypeError, setattr, parser, 'comment_tokens', 123)
        self.assertRaises(TypeError, setattr, parser,
                          'default_start_index', 'abc')
        self.assertRaises(TypeError, setattr, parser, 'sparse_arrays', 'abc')
        self.assertRaises(TypeError, setattr, parser,
                          'global_start_index', 'abc')
        self.assertRaises(TypeError, setattr, parser, 'row_major', 'abc')
        self.assertRaises(TypeError, setattr, parser, 'strict_logical', 'abc')

    def test_float(self):
        test_nml = f90nml.read('float.nml')
        self.assertEqual(self.float_nml, test_nml)
        self.assert_write(test_nml, 'float_target.nml')

    def test_string(self):
        test_nml = f90nml.read('string.nml')
        self.assertEqual(self.string_nml, test_nml)
        self.assert_write(test_nml, 'string_target.nml')

    def test_string_multiline(self):
        test_nml = f90nml.read('string_multiline.nml')
        self.assertEqual(self.string_multiline_nml, test_nml)

    def test_dtype(self):
        test_nml = f90nml.read('dtype.nml')
        self.assertEqual(self.dtype_nml, test_nml)
        self.assert_write(test_nml, 'dtype_target.nml')

    def test_ieee(self):
        test_nml = f90nml.read('ieee.nml')

        # NaN values cannot be tested for equality, so explicitly test values
        self.assertTrue(set(test_nml) == set(self.ieee_nml))
        for grp in test_nml:
            self.assertTrue(set(test_nml[grp]) == set(self.ieee_nml[grp]))
            for key in test_nml[grp]:
                ieee_grp, test_grp = self.ieee_nml[grp], test_nml[grp]
                if key.endswith('_nan'):
                    self.assertTrue(math.isnan(test_grp[key]))
                else:
                    self.assertTrue(ieee_grp[key], test_grp[key])

        self.assert_write(test_nml, 'ieee_target.nml')

    def test_dtype_case(self):
        test_nml = f90nml.read('dtype_case.nml')
        self.assertEqual(self.dtype_case_nml, test_nml)
        self.assert_write(test_nml, 'dtype_case_target.nml')

    def test_bcast(self):
        test_nml = f90nml.read('bcast.nml')
        self.assertEqual(self.bcast_nml, test_nml)
        self.assert_write(test_nml, 'bcast_target.nml')

    def test_comment(self):
        test_nml = f90nml.read('comment.nml')
        self.assertEqual(self.comment_nml, test_nml)
        self.assert_write(test_nml, 'comment_target.nml')

    def test_comment_alt(self):
        parser = f90nml.Parser()
        parser.comment_tokens = '#'
        test_nml = parser.read('comment_alt.nml')
        self.assertEqual(self.comment_alt_nml, test_nml)

    def test_grp_repeat(self):
        test_nml = f90nml.read('grp_repeat.nml')
        self.assertEqual(self.grp_repeat_nml, test_nml)
        self.assert_write(test_nml, 'grp_repeat_target.nml')

    def test_f77(self):
        test_nml = f90nml.read('f77.nml')
        self.assertEqual(self.f77_nml, test_nml)
        self.assert_write(test_nml, 'f77_target.nml')

    def test_dollar(self):
        test_nml = f90nml.read('dollar.nml')
        self.assertEqual(self.dollar_nml, test_nml)
        self.assert_write(test_nml, 'dollar_target.nml')

    def test_multiline(self):
        test_nml = f90nml.read('multiline.nml')
        self.assertEqual(self.multiline_nml, test_nml)
        self.assert_write(test_nml, 'multiline.nml')

    def test_multiline_index(self):
        test_nml = f90nml.read('multiline_index.nml')
        self.assertEqual(self.multiline_nml, test_nml)
        self.assert_write(test_nml, 'multiline_index.nml')

    def test_long_varname(self):
        test_nml = f90nml.read('types.nml')
        test_nml.column_width = 10
        self.assert_write(test_nml, 'types_col10.nml')

    def test_long_string(self):
        test_nml = f90nml.read('long_string.nml')
        self.assertEqual(self.long_string_nml, test_nml)
        test_nml.split_strings = True
        self.assert_write(test_nml, 'long_string_target.nml')

        self.assertRaises(TypeError, setattr, test_nml, 'split_strings', 123)

    def test_ext_token(self):
        test_nml = f90nml.read('ext_token.nml')
        self.assertEqual(self.ext_token_nml, test_nml)

    def test_write_existing_file(self):
        tmp_fname = 'tmp.nml'
        open(tmp_fname, 'w').close()
        test_nml = f90nml.read('empty.nml')
        self.assertRaises(IOError, test_nml.write, tmp_fname)
        os.remove(tmp_fname)

    def test_pop_key(self):
        test_nml = f90nml.read('empty.nml')
        test_nml.pop('empty_nml')
        self.assertEqual(test_nml, f90nml.namelist.Namelist())

    def test_patch_paths(self):
        patch_nml = f90nml.read('types_patch.nml')
        f90nml.patch('types.nml', patch_nml, 'tmp.nml')
        test_nml = f90nml.read('tmp.nml')
        try:
            self.assertEqual(test_nml, patch_nml)
        finally:
            os.remove('tmp.nml')

    def test_patch_files(self):
        patch_nml = f90nml.read('types_patch.nml')
        with open('types.nml') as f_in:
            with open('tmp.nml', 'w') as f_out:
                f90nml.patch(f_in, patch_nml, f_out)
                self.assertFalse(f_in.closed)
                self.assertFalse(f_out.closed)
        try:
            test_nml = f90nml.read('tmp.nml')
            self.assertEqual(test_nml, patch_nml)
        finally:
            os.remove('tmp.nml')

    def test_patch_case(self):
        patch_nml = f90nml.read('types_patch.nml')
        f90nml.patch('types_uppercase.nml', patch_nml, 'tmp.nml')
        test_nml = f90nml.read('tmp.nml')
        try:
            self.assertEqual(test_nml, patch_nml)
        finally:
            os.remove('tmp.nml')

    def test_patch_typeerror(self):
        self.assertRaises(TypeError, f90nml.patch, 'types.nml', 'xyz',
                          'tmp.nml')

    def test_patch_list(self):
        f90nml.patch('list_patch.nml', self.list_patch_nml, 'tmp.nml')
        test_nml = f90nml.read('tmp.nml')
        try:
            self.assertEqual(test_nml, self.list_patch_nml)
        finally:
            os.remove('tmp.nml')

    def test_repatch(self):
        f90nml.patch('repatch.nml', self.repatch_nml, 'tmp.nml')
        test_nml = f90nml.read('tmp.nml')
        try:
            self.assertEqual(test_nml, self.repatch_nml)
        finally:
            os.remove('tmp.nml')

    def test_default_patch(self):
        patch_nml = f90nml.read('types_patch.nml')
        f90nml.patch('types.nml', patch_nml)
        test_nml = f90nml.read('types.nml~')
        try:
            self.assertEqual(test_nml, patch_nml)
        finally:
            os.remove('types.nml~')

        # The above behavior is only for paths, not files
        with open('types.nml') as nml_file:
            self.assertRaises(ValueError, f90nml.patch, nml_file, patch_nml)

    def test_patch_null(self):
        try:
            f90nml.patch('types.nml', {}, 'tmp.nml')
            self.assert_file_equal('types.nml', 'tmp.nml')
        finally:
            os.remove('tmp.nml')

    def test_no_selfpatch(self):
        patch_nml = f90nml.read('types_patch.nml')
        self.assertRaises(ValueError, f90nml.patch,
                          'types.nml', patch_nml, 'types.nml')

    def test_comment_patch(self):
        nml = {'comment_nml': {'v_cmt_inline': 456}}
        try:
            f90nml.patch('comment.nml', nml, 'tmp.nml')
            self.assert_file_equal('comment_patch.nml', 'tmp.nml')
        finally:
            os.remove('tmp.nml')

    def test_parser_default_index(self):
        parser = f90nml.Parser()

        parser.default_start_index = 1
        test_nml = parser.read('default_index.nml')
        self.assertEqual(self.default_one_index_nml, test_nml)

        parser.default_start_index = 0
        test_nml = parser.read('default_index.nml')
        self.assertEqual(self.default_zero_index_nml, test_nml)

    def test_global_index(self):
        parser = f90nml.Parser()
        parser.global_start_index = 1
        test_nml = parser.read('global_index.nml')
        self.assertEqual(self.global_index_nml, test_nml)

    def test_namelist_default_index(self):
        d = {'x_nml': {'x': [1, 2, 3]}}
        test_nml = f90nml.Namelist(d, default_start_index=1)
        # TODO: Check value

    def test_index_syntax(self):
        self.assertRaises(ValueError, f90nml.read, 'index_empty.nml')
        self.assertRaises(ValueError, f90nml.read, 'index_bad.nml')
        self.assertRaises(ValueError, f90nml.read, 'index_bad_start.nml')
        self.assertRaises(ValueError, f90nml.read, 'index_empty_end.nml')
        self.assertRaises(ValueError, f90nml.read, 'index_bad_end.nml')
        self.assertRaises(ValueError, f90nml.read, 'index_empty_stride.nml')
        self.assertRaises(ValueError, f90nml.read, 'index_bad_stride.nml')
        self.assertRaises(ValueError, f90nml.read, 'index_zero_stride.nml')

    def test_bad_start_index(self):
        nml = Namelist()
        self.assertRaises(TypeError, setattr, nml, 'start_index', 'abc')
        self.assertRaises(TypeError, setattr, nml, 'default_start_index',
                          'abc')

    def test_iter_in_getitem(self):
        d = {'a': {'b': 1.}}
        nml = Namelist(d)
        self.assertEqual(nml[('a', 'b')], 1.)
        self.assertEqual(nml[['a', 'b']], 1.)
        self.assertEqual(nml['a']['b'], 1.)

    def test_groups(self):
        d = {'a': {'b': 1.}}
        nml = Namelist(d)
        key, value = next(nml.groups())
        self.assertEqual(key, ('a', 'b'))
        self.assertEqual(value, 1.)

    def test_f90repr(self):
        nml = Namelist()
        self.assertEqual(nml._f90repr(1), '1')
        self.assertEqual(nml._f90repr(1.), '1.0')
        self.assertEqual(nml._f90repr(1+2j), '(1.0, 2.0)')
        self.assertEqual(nml._f90repr(True), '.true.')
        self.assertEqual(nml._f90repr(False), '.false.')
        self.assertEqual(nml._f90repr('abc'), "'abc'")

        for ptype in ({}, [], set()):
            self.assertRaises(ValueError, nml._f90repr, ptype)

    def test_pybool(self):
        for fstr_true in ('true', '.true.', 't', '.t.'):
            self.assertEqual(pybool(fstr_true), True)

        for fstr_false in ('false', '.false.', 'f', '.f.'):
            self.assertEqual(pybool(fstr_false), False)

        for fstr_true in ('ture', '.t'):
            self.assertEqual(pybool(fstr_true, strict_logical=False), True)

        for fstr_false in ('flase', '.f'):
            self.assertEqual(pybool(fstr_false, strict_logical=False), False)

        for fstr in ('ture', '.t', 'flase', '.f'):
            self.assertRaises(ValueError, pybool, fstr)

        for fstr in ('g', '.', 'xyz'):
            self.assertRaises(ValueError, pybool, fstr, strict_logical=False)

    def test_close_patch_on_error(self):
        patch = {'tmp_nml': {'tmp_val': 0}}
        self.assertRaises(ValueError, f90nml.patch, 'index_empty.nml', patch,
                                                    'tmp.nml')
        os.remove('tmp.nml')

    def test_indent(self):
        test_nml = f90nml.read('types.nml')

        test_nml.indent = 2
        self.assert_write(test_nml, 'types_indent_2.nml')

        test_nml.indent = '\t'
        self.assert_write(test_nml, 'types_indent_tab.nml')

        self.assertRaises(ValueError, setattr, test_nml, 'indent', -4)
        self.assertRaises(ValueError, setattr, test_nml, 'indent', 'xyz')
        self.assertRaises(TypeError, setattr, test_nml, 'indent', [1, 2, 3])

    def test_column_width(self):
        test_nml = f90nml.read('multiline.nml')
        test_nml.column_width = 40
        self.assert_write(test_nml, 'multiline_colwidth.nml')

        self.assertRaises(ValueError, setattr, test_nml, 'column_width', -1)
        self.assertRaises(TypeError, setattr, test_nml, 'column_width', 'xyz')

    def test_end_comma(self):
        test_nml = f90nml.read('types.nml')
        test_nml.end_comma = True
        self.assert_write(test_nml, 'types_end_comma.nml')

        self.assertRaises(TypeError, setattr, test_nml, 'end_comma', 'xyz')

    def test_uppercase(self):
        test_nml = f90nml.read('types.nml')
        test_nml.uppercase = True
        self.assert_write(test_nml, 'types_uppercase.nml')

        self.assertRaises(TypeError, setattr, test_nml, 'uppercase', 'xyz')

    def test_index_spacing(self):
        test_nml = f90nml.read('multidim.nml')
        test_nml.index_spacing = True

        self.assertEqual(test_nml.index_spacing, True)
        self.assert_write(test_nml, 'multidim_idx_space.nml')

        self.assertRaises(TypeError, setattr, test_nml, 'index_spacing', 123)

    def test_float_format(self):
        test_nml = f90nml.read('float.nml')
        test_nml.float_format = '.3f'
        self.assert_write(test_nml, 'float_format.nml')

        self.assertRaises(TypeError, setattr, test_nml, 'float_format', 123)

    def test_logical_repr(self):
        parser = f90nml.Parser()
        parser.strict_logical = False
        test_nml = parser.read('logical.nml')
        test_nml.true_repr = 'T'
        test_nml.false_repr = 'F'

        self.assertEqual(test_nml.false_repr, test_nml.logical_repr[0])
        self.assertEqual(test_nml.true_repr, test_nml.logical_repr[1])
        self.assert_write(test_nml, 'logical_repr.nml')

        test_nml.logical_repr = 'F', 'T'
        self.assert_write(test_nml, 'logical_repr.nml')

        self.assertRaises(TypeError, setattr, test_nml, 'true_repr', 123)
        self.assertRaises(TypeError, setattr, test_nml, 'false_repr', 123)
        self.assertRaises(ValueError, setattr, test_nml, 'true_repr', 'xyz')
        self.assertRaises(ValueError, setattr, test_nml, 'false_repr', 'xyz')
        self.assertRaises(TypeError, setattr, test_nml, 'logical_repr', 'xyz')
        self.assertRaises(ValueError, setattr, test_nml, 'logical_repr', [])

    def test_findex_iteration(self):
        rng = [(None, 5, None)]
        fidx = iter(FIndex(rng))

        for i, j in enumerate(fidx, start=1):
            self.assertEqual(i, j[0])

    def test_dict_write(self):
        self.assert_write(self.types_nml, 'types_dict.nml')

    def test_dict_assign(self):
        test_nml = f90nml.Namelist()
        test_nml['dict_group'] = {'a': 1, 'b': 2}
        try:
            test_nml.write('tmp.nml')
        finally:
            os.remove('tmp.nml')

    def test_winfmt(self):
        test_nml = f90nml.read('winfmt.nml')
        self.assertEqual(self.winfmt_nml, test_nml)

    def test_eof_no_cr(self):
        test_nml = f90nml.read('no_eol_in_eof.nml')
        # TODO: Test values

    def test_namelist_patch(self):
        nml = f90nml.Namelist({
            'a_nml': {
                'x': 1,
                'y': 2,
            }
        })

        # Check overwriting values
        nml.patch({'a_nml': {'x': 3}})

        self.assertEqual(nml['a_nml']['x'], 3)
        self.assertEqual(nml['a_nml']['y'], 2)

        # Check appending values doesn't remove previous
        nml.patch({'a_nml': {'z': 5}})

        self.assertEqual(nml['a_nml']['x'], 3)
        self.assertEqual(nml['a_nml']['y'], 2)
        self.assertEqual(nml['a_nml']['z'], 5)

        # Check adding a new section also works
        nml.patch({
            'b_nml': {'q': 33},
            'a_nml': {'z': 4}
        })

        self.assertEqual(nml['a_nml']['z'], 4)
        self.assertEqual(nml['b_nml']['q'], 33)
        self.assertRaises(KeyError, nml['b_nml'].__getitem__, 'z')

    def test_sorted_output(self):
        test_nml = f90nml.read('types.nml')
        self.assert_write(test_nml, 'types_sorted.nml', sort=True)

    def test_extern_cmt(self):
        test_nml = f90nml.read('extern_cmt.nml')
        self.assertEqual(self.extern_cmt_nml, test_nml)

    def test_print_nml(self):
        nml = f90nml.read('types.nml')

        stdout = StringIO()
        print(nml, file=stdout)
        stdout.seek(0)
        source_str = stdout.read()
        stdout.close()

        with open('types.nml') as target:
            target_str = target.read()

        self.assertEqual(source_str, target_str)

    def test_print_group(self):
        nml = f90nml.read('types.nml')

        stdout = StringIO()
        print(nml['types_nml'], file=stdout)
        stdout.seek(0)
        source_str = stdout.read().rstrip('\n')
        stdout.close()

        target_str = repr(nml['types_nml'])

        self.assertEqual(source_str, target_str)

    def test_gen_dtype(self):
        d = {'dtype_nml': {'a': [{'b': 1, 'c': 2}, {'b': 3, 'c': 4}]}}
        nml = f90nml.Namelist(d)
        out = StringIO()
        print(nml, file=out)
        # TODO: Check output
        out.close()

    def test_gen_multidim(self):
        d = {'md_nml': {'x': [[1, 2, 3], [4, 5, 6], [7, 8, 9]]}}
        nml = f90nml.Namelist(d)
        out = StringIO()
        print(nml, file=out)
        # TODO: Check output
        out.close()

    if has_numpy:
        def test_numpy_write(self):
            self.assert_write(self.numpy_nml, 'numpy_types.nml')

    def test_read_string(self):
        test_nml = f90nml.reads(
            '&types_nml'
            ' v_integer = 1'
            ' v_float = 1.0'
            ' v_complex = (1.0, 2.0)'
            ' v_logical = .true.'
            ' v_string = \'Hello\''
            '/'
        )
        self.assertEqual(self.types_nml, test_nml)

    # CLI tests
    def test_cli_help(self):
        cmd = ['f90nml']
        self.get_cli_output(cmd)

    def test_cli_read(self):
        cmd = ['f90nml', 'types.nml']
        source_str = self.get_cli_output(cmd)

        with open('types.nml') as target:
            target_str = target.read()
            self.assertEqual(source_str, target_str)

    def test_cli_gen(self):
        cmd = ['f90nml', '-g', 'gen_nml', '-v', 'x=1']
        source_str = self.get_cli_output(cmd)

        with open('gen.nml') as target:
            target_str = target.read()

        self.assertEqual(source_str, target_str)

    def test_cli_replace(self):
        cmd = ['f90nml', '-g', 'types_nml', '-v', 'v_integer=5',
               '-v', 'v_logical=.false.', 'types.nml']
        source_str = self.get_cli_output(cmd)

        with open('types_cli_set.nml') as target:
            target_str = target.read()

        self.assertEqual(source_str, target_str)

    def test_cli_replace_no_group(self):
        cmd = ['f90nml', '-v', 'v_integer=5', '-v', 'v_logical=.false.',
               'types.nml']
        source_str = self.get_cli_output(cmd)

        # TODO: Check stderr
        error_str = ("f90nml: warning: Assuming variables are in group "
                     "'types_nml'.\n")

        with open('types_cli_set.nml') as target:
            target_str = target.read()
            self.assertEqual(source_str, target_str)

    def test_cli_replace_write(self):
        cmd = ['f90nml', '-g', 'types_nml', '-v', 'v_integer=5',
               '-v', 'v_logical=.false.', 'types.nml', 'tmp.nml']
        self.get_cli_output(cmd)

        with open('tmp.nml') as source:
            with open('types_cli_set.nml') as target:
                source_str = source.read()
                target_str = target.read()

        self.assertEqual(source_str, target_str)

    def test_cli_patch(self):
        cmd = ['f90nml', '-p', '-g', 'comment_nml', '-v', 'v_cmt_inline=456',
               'comment.nml']
        source_str = self.get_cli_output(cmd)

        with open('comment_patch.nml') as target:
            target_str = target.read()

        self.assertEqual(source_str, target_str)

    def test_cli_bad_format(self):
        cmd = ['f90nml', '-f', 'blah', 'types.nml']
        source_str = self.get_cli_output(cmd, get_stderr=True)
        # TODO: Automate the format list
        target_str = ("f90nml: error: format must be one of the following: "
                      "('json', 'yaml', 'nml')\n")

        self.assertEqual(source_str, target_str)

    def test_cli_json_write(self):
        cmd = ['f90nml', 'types.nml', 'tmp.json']
        out = self.get_cli_output(cmd)

        self.assert_file_equal('types.json', 'tmp.json')

        # TODO: These are just throwaway tests that need to be moved after the
        # sorting issues have been sorted
        cmd = ['f90nml', 'vector.nml', 'tmp.json']
        out = self.get_cli_output(cmd)

        cmd = ['f90nml', 'dtype.nml', 'tmp.json']
        out = self.get_cli_output(cmd)

        os.remove('tmp.json')

    def test_cli_json_fmt(self):
        cmd = ['f90nml', '-f', 'json', 'types.nml']
        source_str = self.get_cli_output(cmd)

        with open('types.json') as target:
            target_str = target.read()

        self.assertEqual(source_str, target_str)

    def test_cli_json_read(self):
        cmd = ['f90nml', 'types.json']
        source_str = self.get_cli_output(cmd)
        # TODO: Check output after resolving the ordering issue

        # Quickly test some of the list features
        cmd = ['f90nml', 'vector.json']
        source_str = self.get_cli_output(cmd)

    def test_cli_json_patch_fail(self):
        error_str = 'f90nml: error: Only namelist files can be patched.\n'

        # JSON input patch
        cmd = ['f90nml', '-p', '-v', 'steps=432', 'types.json']
        source_str = self.get_cli_output(cmd, get_stderr=True)
        self.assertEqual(source_str, error_str)

        # JSON output patch
        cmd = ['f90nml', '-p', '-v', 'steps=432', 'config.nml', 'tmp.json']
        source_str = self.get_cli_output(cmd, get_stderr=True)
        self.assertEqual(source_str, error_str)

    if has_yaml:
        def test_cli_yaml_write(self):
            cmd = ['f90nml', 'types.nml', 'tmp.yaml']
            out = self.get_cli_output(cmd)

            self.assert_file_equal('types.yaml', 'tmp.yaml')
            os.remove('tmp.yaml')

        def test_cli_yaml_read(self):
            cmd = ['f90nml', 'types.yaml']
            source_str = self.get_cli_output(cmd)
            # TODO: Check output after resolving the ordering issue

    def test_cli_missing_yaml(self):
        orig_has_yaml = f90nml.cli.has_yaml
        f90nml.cli.has_yaml = False

        cmd = ['f90nml', 'types.yaml']
        source_str = self.get_cli_output(cmd, get_stderr=True)

        target_str = (
            'f90nml: error: YAML module could not be found.\n'
            '  To enable YAML support, install PyYAML or use the f90nml[yaml] '
            'package.\n'
        )
        self.assertEqual(source_str, target_str)

        f90nml.cli.has_yaml = orig_has_yaml

    def test_check_repeat_flag(self):
        nml = f90nml.Namelist()
        self.assertFalse(nml.repeat_counter)

    def test_set_repeat_flag(self):
        nml = f90nml.Namelist()
        nml.repeat_counter = True
        self.assertTrue(nml.repeat_counter)

    def test_set_repeat_flag_incorrect(self):
        nml = f90nml.Namelist()
        with self.assertRaises(TypeError):
            nml.repeat_counter = 'Hello'

    # TODO fails in _var_strings
    # def test_repeat_empty(self):
        # nml_dict = {'a':{'b' : []}}
        # nml = f90nml.Namelist(nml_dict)
        # nml.repeat = True
        # out = StringIO()
        # print(nml, file=out)
        # out.seek(0)
        # line1 = out.readline()
        # line2 = out.readline()
        # self.assertEqual(line2.lstrip(), 'b = 1, 2, 3\n')

    def test_repeat_scalar(self):
        nml = f90nml.Namelist({'a': {'b': 1}})
        nml.repeat_counter = True
        out = StringIO()
        print(nml, file=out)
        out.seek(0)
        line1 = out.readline()
        line2 = out.readline()
        self.assertEqual(line2.lstrip(), 'b = 1\n')

    def test_repeat_single(self):
        nml_dict = {'a': {'b': [1]}}
        nml = f90nml.Namelist(nml_dict)
        nml.repeat_counter = True
        out = StringIO()
        print(nml, file=out)
        out.seek(0)
        line1 = out.readline()
        line2 = out.readline()
        self.assertEqual(line2.lstrip(), 'b = 1\n')

    def test_repeat_no_consecutive(self):
        nml_dict = {'a': {'b': [1, 2, 3]}}
        nml = f90nml.Namelist(nml_dict)
        nml.repeat_counter = True
        out = StringIO()
        print(nml, file=out)
        out.seek(0)
        line1 = out.readline()
        line2 = out.readline()
        self.assertEqual(line2.lstrip(), 'b = 1, 2, 3\n')

    def test_repeat_middle_consecutive(self):
        nml_dict = {'a': {'b': [1, 2, 2, 3]}}
        nml = f90nml.Namelist(nml_dict)
        nml.repeat_counter = True
        out = StringIO()
        print(nml, file=out)
        out.seek(0)
        line1 = out.readline()
        line2 = out.readline()
        self.assertEqual(line2.lstrip(), 'b = 1, 2*2, 3\n')

    def test_repeat_beginning_consecutive(self):
        nml_dict = {'a': {'b': [1, 1, 2, 3]}}
        nml = f90nml.Namelist(nml_dict)
        nml.repeat_counter = True
        out = StringIO()
        print(nml, file=out)
        out.seek(0)
        line1 = out.readline()
        line2 = out.readline()
        self.assertEqual(line2.lstrip(), 'b = 2*1, 2, 3\n')

    def test_repeat_end_consecutive(self):
        nml_dict = {'a': {'b': [1, 2, 3, 3]}}
        nml = f90nml.Namelist(nml_dict)
        nml.repeat_counter = True
        out = StringIO()
        print(nml, file=out)
        out.seek(0)
        line1 = out.readline()
        line2 = out.readline()
        self.assertEqual(line2.lstrip(), 'b = 1, 2, 2*3\n')

    def test_repeat_consecutive_repeating(self):
        nml_dict = {'a': {'b': [1, 1, 2, 3, 1, 1, 1]}}
        nml = f90nml.Namelist(nml_dict)
        nml.repeat_counter = True
        out = StringIO()
        print(nml, file=out)
        out.seek(0)
        line1 = out.readline()
        line2 = out.readline()
        self.assertEqual(line2.lstrip(), 'b = 2*1, 2, 3, 3*1\n')

    def test_repeat_all_repeating(self):
        nml_dict = {'a': {'b': [1, 1, 1, 1, 1]}}
        nml = f90nml.Namelist(nml_dict)
        nml.repeat_counter = True
        out = StringIO()
        print(nml, file=out)
        out.seek(0)
        line1 = out.readline()
        line2 = out.readline()
        self.assertEqual(line2.lstrip(), 'b = 5*1\n')

    def test_repeat_repeating_float(self):
        nml_dict = {'a': {'b': [1.0, 1.0, 2.0, 0.3]}}
        nml = f90nml.Namelist(nml_dict)
        nml.repeat_counter = True
        out = StringIO()
        print(nml, file=out)
        out.seek(0)
        line1 = out.readline()
        line2 = out.readline()
        self.assertEqual(line2.lstrip(), 'b = 2*1.0, 2.0, 0.3\n')

    def test_repeat_repeating_complex(self):
        nml_dict = {'a': {'b': [1+2j, 1+2j, 3j]}}
        nml = f90nml.Namelist(nml_dict)
        nml.repeat_counter = True
        out = StringIO()
        print(nml, file=out)
        out.seek(0)
        line1 = out.readline()
        line2 = out.readline()
        self.assertEqual(line2.lstrip(), 'b = 2*(1.0, 2.0), (0.0, 3.0)\n')

    def test_repeat_repeating_logical(self):
        nml_dict = {'a': {'b': [True, True, False]}}
        nml = f90nml.Namelist(nml_dict)
        nml.repeat_counter = True
        out = StringIO()
        print(nml, file=out)
        out.seek(0)
        line1 = out.readline()
        line2 = out.readline()
        self.assertEqual(line2.lstrip(), "b = 2*.true., .false.\n")

    # Failed namelist parsing
    # NOTE: This is a very weak test, since '& x=1' / will pass
    def test_grp_token_end(self):
        self.assertRaises(ValueError, f90nml.reads, '&')

    def test_no_group_name(self):
        self.assertRaises(ValueError, f90nml.reads, '& /')

    def test_empty_grp_no_end(self):
        self.assertRaises(ValueError, f90nml.reads, '&end')

    def test_string_grp_no_end(self):
        self.assertRaises(ValueError, f90nml.reads, '&a_nml x = 1')

    def test_file_grp_no_end(self):
        self.assertRaises(ValueError, f90nml.read, 'grp_no_end.nml')


if __name__ == '__main__':
    if os.path.isfile('tmp.nml'):
        os.remove('tmp.nml')
    unittest.main()
